# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# To avoid possible naming conflict, mars functions and classes should be renamed.
# Functions should be renamed by adding a prefix 'mars_', and classes should be renamed
# by adding a prefix 'Mars'.

import functools
import inspect
import warnings
from abc import ABC, abstractmethod
from collections import defaultdict
from types import ModuleType
from typing import Any, Callable, Dict, Generator, List, Optional, Tuple, Type, Union

# For maintenance, any module wants to import from mars, it should import from here.
from .._mars import dataframe as mars_dataframe
from .._mars import execute as mars_execute
from .._mars import new_session as mars_new_session
from .._mars import remote as mars_remote
from .._mars import stop_server as mars_stop_server
from .._mars import tensor as mars_tensor
from .._mars.core import Entity as MarsEntity
from .._mars.core import OutputType as MarsOutputType
from .._mars.core.entity.executable import ExecutableTuple
from .._mars.core.entity.objects import OBJECT_TYPE as MARS_OBJECT_TYPE
from .._mars.dataframe import DataFrame as MarsDataFrame
from .._mars.dataframe import Index as MarsIndex
from .._mars.dataframe import Series as MarsSeries
from .._mars.dataframe.base.accessor import CachedAccessor as MarsCachedAccessor
from .._mars.dataframe.base.accessor import DatetimeAccessor as MarsDatetimeAccessor
from .._mars.dataframe.base.accessor import StringAccessor as MarsStringAccessor
from .._mars.dataframe.core import CATEGORICAL_TYPE as MARS_CATEGORICAL_TYPE
from .._mars.dataframe.core import DATAFRAME_GROUPBY_TYPE as MARS_DATAFRAME_GROUPBY_TYPE
from .._mars.dataframe.core import (
    DATAFRAME_OR_SERIES_TYPE as MARS_DATAFRAME_OR_SERIES_TYPE,
)
from .._mars.dataframe.core import DATAFRAME_TYPE as MARS_DATAFRAME_TYPE
from .._mars.dataframe.core import INDEX_TYPE as MARS_INDEX_TYPE
from .._mars.dataframe.core import SERIES_GROUPBY_TYPE as MARS_SERIES_GROUPBY_TYPE
from .._mars.dataframe.core import SERIES_TYPE as MARS_SERIES_TYPE
from .._mars.dataframe.core import DataFrameGroupBy as MarsDataFrameGroupBy
from .._mars.dataframe.core import SeriesGroupBy as MarsSeriesGroupBy
from .._mars.dataframe.datastore.to_csv import DataFrameToCSV as MarsDataFrameToCSV
from .._mars.dataframe.datastore.to_parquet import (
    DataFrameToParquet as MarsDataFrameToParquet,
)
from .._mars.dataframe.datastore.to_sql import (
    DataFrameToSQLTable as MarsDataFrameToSQLTable,
)
from .._mars.dataframe.datastore.to_vineyard import (
    DataFrameToVineyardChunk as MarsDataFrameToVineyardChunk,
)
from .._mars.dataframe.indexing.at import DataFrameAt as MarsDataFrameAt
from .._mars.dataframe.indexing.iat import DataFrameIat as MarsDataFrameIat
from .._mars.dataframe.indexing.iloc import DataFrameIloc as MarsDataFrameIloc
from .._mars.dataframe.indexing.loc import DataFrameLoc as MarsDataFrameLoc
from .._mars.dataframe.plotting.core import PlotAccessor as MarsPlotAccessor
from .._mars.dataframe.window.ewm.core import EWM as MarsEWM
from .._mars.dataframe.window.expanding.core import Expanding as MarsExpanding
from .._mars.dataframe.window.rolling.core import Rolling as MarsRolling
from .._mars.deploy.oscar import session
from .._mars.tensor.core import TENSOR_TYPE as MARS_TENSOR_TYPE
from .._mars.tensor.core import Tensor as MarsTensor
from .._mars.tensor.core import flatiter as mars_flatiter
from .._mars.tensor.lib.index_tricks import CClass as MarsCClass
from .._mars.tensor.lib.index_tricks import MGridClass as MarsMGridClass
from .._mars.tensor.lib.index_tricks import OGridClass as MarsOGridClass
from .._mars.tensor.lib.index_tricks import RClass as MarsRClass
from .data import DATA_MEMBERS, Data, DataRef, DataType
from .utils.docstring import attach_cls_member_docstring


def own_data(mars_entity: MarsEntity) -> bool:
    # There are several mars operands which holds data directly. For example,
    # `DataFrameDataSource`, it is an operand represents creating a DataFrame
    # from Pandas DataFrame and its member `data` is the Pandas DataFrame. For
    # those mars entities that created by operands like these can skip executions
    # when users iterate or print entities. This function is to check if
    # mars entity's operand owns data.
    from .._mars.dataframe.datasource.dataframe import DataFrameDataSource
    from .._mars.dataframe.datasource.index import IndexDataSource
    from .._mars.dataframe.datasource.series import SeriesDataSource
    from .._mars.tensor.datasource import ArrayDataSource

    if (
        isinstance(
            mars_entity.op,
            (ArrayDataSource, DataFrameDataSource, SeriesDataSource, IndexDataSource),
        )
        and mars_entity.op.data is not None
    ):
        return True
    return False


# mars class name -> execution conditions
_TO_MARS_EXECUTION_CONDITION: Dict[
    str, List[Callable[["MarsEntity"], bool]]
] = defaultdict(list)
_FROM_MARS_EXECUTION_CONDITION: Dict[
    str, List[Callable[["MarsEntity"], bool]]
] = defaultdict(list)


def register_to_mars_execution_condition(
    mars_entity_type: str, condition: Callable[["MarsEntity"], bool]
):
    _TO_MARS_EXECUTION_CONDITION[mars_entity_type].append(condition)


def register_from_mars_execution_condition(
    mars_entity_type: str, condition: Callable[["MarsEntity"], bool]
):
    _FROM_MARS_EXECUTION_CONDITION[mars_entity_type].append(condition)


_MARS_CLS_TO_CONVERTER: Dict[Type, Callable] = {}


def register_converter(from_cls_list: List[Type]):
    """
    A decorator for convenience of registering a class converter.
    """

    def decorate(cls: Type):
        for from_cls in from_cls_list:
            assert from_cls not in _MARS_CLS_TO_CONVERTER
            _MARS_CLS_TO_CONVERTER[from_cls] = cls
        return cls

    return decorate


class ClsMethodWrapper(ABC):
    def __init__(
        self,
        func_name: str = "",
        library_cls: Type = object,
        fallback_warning: bool = False,
    ):
        self.library_cls = library_cls
        self.func_name = func_name
        self.fallback_warning = fallback_warning

    @abstractmethod
    def _generate_fallback_data(self, mars_entity: MarsEntity) -> Any:
        """
        let mars entity fallback to data according to the library

        Parameters
        ----------
        mars_entity: MarsEntity

        Returns
        -------

        """

    @abstractmethod
    def _generate_warning_msg(self, mars_entity: MarsEntity, func_name: str) -> str:
        """
        generate fallback warning message according to the library

        Parameters
        ----------
        mars_entity: MarsEntity
        func_name: str

        Returns
        -------
        warning_msg: str
        """

    @abstractmethod
    def _get_output_type(self, func: Callable) -> MarsOutputType:
        """
        get output type according to the library

        Parameters
        ----------
        func: Callable

        Returns
        -------
        output_type: MarsOutputType
        """

    @abstractmethod
    def _get_docstring_src_module(self) -> ModuleType:
        """
        get docstring src module according to the library
        """

    def get_wrapped(self) -> Callable:
        """
        wrap pd.DataFrame member functions, np.ndarray methods, and other methods

        returns a callable
        """

        @functools.wraps(getattr(self.library_cls, self.func_name))
        def _wrapped(entity: MarsEntity, *args, **kwargs):
            def _spawn(entity: MarsEntity) -> MarsEntity:
                """
                Execute pandas/numpy fallback with mars remote.
                """

                def execute_func(
                    mars_entity: MarsEntity, f_name: str, *args, **kwargs
                ) -> Any:
                    ret = self._generate_fallback_data(mars_entity)
                    return getattr(ret, f_name)(*args, **kwargs)

                new_args = (entity, self.func_name) + args
                ret = mars_remote.spawn(
                    execute_func, args=new_args, kwargs=kwargs, output_types="object"
                )
                return from_mars(ret.execute())

            def _map_chunk(entity: MarsEntity, skip_infer: bool = False) -> MarsEntity:
                """
                Execute pandas fallback with map_chunk.
                """
                ret = entity.map_chunk(
                    lambda x, *args, **kwargs: getattr(x, self.func_name)(
                        *args, **kwargs
                    ),
                    args=args,
                    kwargs=kwargs,
                    skip_infer=skip_infer,
                )
                if skip_infer:
                    ret = ret.ensure_data()
                return from_mars(ret)

            warnings.warn(
                self._generate_warning_msg(entity, self.func_name),
                RuntimeWarning,
            )

            # rechunk mars tileable as one chunk
            one_chunk_entity = entity.rechunk(max(entity.shape))

            if hasattr(one_chunk_entity, "map_chunk"):
                try:
                    return _map_chunk(one_chunk_entity, skip_infer=False)
                except TypeError:
                    # when infer failed in map_chunk, we would use remote to execute
                    # or skip inferring
                    output_type = self._get_output_type(
                        getattr(self.library_cls, self.func_name)
                    )
                    if output_type == MarsOutputType.object:
                        return _spawn(one_chunk_entity)
                    else:
                        # skip_infer = True to avoid TypeError raised by inferring
                        return _map_chunk(one_chunk_entity, skip_infer=True)
            else:
                return _spawn(one_chunk_entity)

        attach_cls_member_docstring(
            _wrapped,
            self.func_name,
            docstring_src_module=self._get_docstring_src_module(),
            docstring_src_cls=self.library_cls,
            fallback_warning=self.fallback_warning,
        )
        return _wrapped


def wrap_magic_method(method_name: str) -> Callable[[Any], Any]:
    def wrapped(self: DataRef, *args, **kwargs):
        # trigger on condition execution.
        mars_entity = to_mars(self)
        if (mars_entity is None) or (
            not hasattr(mars_entity, method_name)
        ):  # pragma: no cover
            raise AttributeError(
                f"'{self.data.data_type.name}' object has no attribute '{method_name}'"
            )
        else:
            return wrap_mars_callable(
                getattr(mars_entity, method_name),
                attach_docstring=False,
                is_cls_member=True,
            )(*args, **kwargs)

    return wrapped


def wrap_generator(wrapped: Generator):
    for item in wrapped:
        yield from_mars(item)


def wrap_member_func(member_func: Callable, mars_entity: MarsEntity):
    @functools.wraps(member_func)
    def _wrapped(*args, **kwargs):
        return member_func(mars_entity, *args, **kwargs)

    return _wrapped


class MemberProxy:
    @classmethod
    def getattr(cls, ref: DataRef, item: str):
        # trigger on condition execution.
        mars_entity = to_mars(ref)
        data_type = ref.data.data_type
        member = DATA_MEMBERS[data_type].get(item, None)
        if member is not None and callable(member):
            ret = wrap_member_func(member, mars_entity)
            ret.__doc__ = member.__doc__
            return ret

        if not hasattr(mars_entity, item):
            raise AttributeError(f"'{data_type.name}' object has no attribute '{item}'")

        attr = getattr(mars_entity, item, None)
        if callable(attr):
            return wrap_mars_callable(
                attr,
                attach_docstring=True,
                is_cls_member=True,
                member_name=item,
                data_type=data_type,
            )
        else:
            # e.g. string accessor
            return from_mars(attr)

    @classmethod
    def setattr(cls, ref: DataRef, key: str, value: Any):
        # trigger on condition execution.
        mars_entity = to_mars(ref)
        if isinstance(getattr(type(mars_entity), key, None), property):
            # call the setter of the specified property.
            getattr(type(mars_entity), key).fset(mars_entity, to_mars(value))
        else:
            mars_entity.__setattr__(key, value)


class MarsGetItemProxy:
    def __init__(self, mars_obj):
        self._mars_obj = mars_obj

    def __getitem__(self, item):
        return from_mars(self._mars_obj[to_mars(item)])


class MarsGetAttrProxy:
    def __init__(self, obj: Any):
        self._mars_obj = to_mars(obj)

    def __getattr__(self, item):
        mars_obj = object.__getattribute__(self, "_mars_obj")
        attr = getattr(mars_obj, item, None)
        if attr is None:
            raise AttributeError(f"no attribute '{item}'")
        elif callable(attr):  # pragma: no cover
            return wrap_mars_callable(attr, attach_docstring=False, is_cls_member=True)
        else:  # pragma: no cover
            # class variable
            return from_mars(attr)

    def __getitem__(self, item):
        return from_mars(self._mars_obj[to_mars(item)])


def to_mars(inp: Union[DataRef, Tuple, List, Dict]):
    """
    Convert xorbits data references to mars entities and execute them if needed.
    """

    if isinstance(inp, DataRef):
        mars_entity = getattr(inp.data, "_mars_entity", None)
        if mars_entity is None:  # pragma: no cover
            raise TypeError(f"Can't convert {inp} to mars entity")
        conditions = _TO_MARS_EXECUTION_CONDITION[type(mars_entity).__name__]
        for cond in conditions:
            if cond(mars_entity):
                from .execution import run

                run(inp)
        return mars_entity
    elif isinstance(inp, (MarsGetItemProxy, MarsGetAttrProxy)):
        # converters.
        return getattr(inp, "_mars_obj")
    elif isinstance(inp, tuple):
        if type(inp) is tuple or isinstance(inp, ExecutableTuple):
            return tuple(to_mars(i) for i in inp)
        else:
            # named tuple
            return type(inp)(*map(to_mars, inp))
    elif isinstance(inp, list):
        # in-place modification of list
        # preserve weak references to list, avoiding access issues
        for i, item in enumerate(inp):
            inp[i] = to_mars(item)
        return inp
    elif isinstance(inp, dict):
        # in-place modification of dict
        # preserve weak references to dict, avoiding access issues
        for k, v in inp.items():
            inp[k] = to_mars(v)
        return inp
    else:
        return inp


def from_mars(inp: Union[MarsEntity, Tuple, List, Dict, None]):
    """
    Convert mars entities to xorbits data references.
    """
    if isinstance(inp, MarsEntity):
        conditions = _FROM_MARS_EXECUTION_CONDITION[type(inp).__name__]
        ret = DataRef(Data.from_mars(inp))
        for cond in conditions:
            if cond(inp):
                from .execution import run

                run(ret)
        return ret
    elif type(inp) in _MARS_CLS_TO_CONVERTER:
        return _MARS_CLS_TO_CONVERTER[type(inp)](inp)
    elif isinstance(inp, tuple):
        if type(inp) is tuple or isinstance(inp, ExecutableTuple):
            return tuple(from_mars(i) for i in inp)
        else:
            # named tuple
            return type(inp)(*map(from_mars, inp))
    elif isinstance(inp, list):
        # in-place modification of list
        # preserve weak references to list, avoiding access issues
        for i, item in enumerate(inp):
            inp[i] = from_mars(item)
        return inp
    elif isinstance(inp, dict):
        # in-place modification of dict
        # preserve weak references to dict, avoiding access issues
        for k, v in inp.items():
            inp[k] = from_mars(v)
        return inp
    elif isinstance(inp, Generator):
        return wrap_generator(inp)
    else:
        return inp


def wrap_mars_callable(
    c: Callable, attach_docstring: bool, is_cls_member: bool, **kwargs
) -> Callable:
    """
    A function wrapper that makes arguments of the wrapped callable be mars compatible types and
    return value be xorbits compatible types.
    """

    @functools.wraps(c)
    def wrapped(*args, **kwargs):
        return from_mars(c(*to_mars(args), **to_mars(kwargs)))

    if attach_docstring:
        if is_cls_member:
            from .utils.docstring import attach_cls_member_docstring

            return attach_cls_member_docstring(wrapped, **kwargs)
        else:
            from .utils.docstring import attach_module_callable_docstring

            return attach_module_callable_docstring(wrapped, **kwargs)
    else:
        # for methods that do not need a docstring, like methods from mars, we need to reset the
        # docstring to prevent users from seeing a mars docstring.
        wrapped.__doc__ = ""
        return wrapped


def collect_cls_members(
    cls: Type,
    data_type: Optional[DataType] = None,
    docstring_src_module: Optional[ModuleType] = None,
    docstring_src_cls: Optional[Type] = None,
) -> Dict[str, Any]:
    cls_members: Dict[str, Any] = {}
    for name, cls_member in inspect.getmembers(cls):
        # Tileable and TileableData object may have functions that have the same names.
        # For example, Index and IndexData both have `copy` function, but they have completely different semantics.
        # Therefore, when the Index's `copy` method has been collected,
        # the method of the same name on IndexData cannot be collected again.
        if cls.__name__.endswith("Data") and name in DATA_MEMBERS[data_type]:  # type: ignore
            continue
        if inspect.isfunction(cls_member) and not name.startswith("_"):
            cls_members[name] = wrap_mars_callable(
                cls_member,
                attach_docstring=True,
                is_cls_member=True,
                member_name=name,
                data_type=data_type,
                docstring_src_module=docstring_src_module,
                docstring_src_cls=docstring_src_cls,
            )
        elif isinstance(cls_member, property):
            fget = cls_member.fget
            if fget is None:  # pragma: no cover
                raise ValueError(f"property {name} does not have a valid fget method.")
            c = wrap_mars_callable(
                fget,
                attach_docstring=True,
                is_cls_member=True,
                member_name=name,
                data_type=data_type,
                docstring_src_module=docstring_src_module,
                docstring_src_cls=docstring_src_cls,
            )
            cls_members[name] = property(c)
            cls_members[name].__doc__ = c.__doc__

    return cls_members


def register_data_members(data_type: DataType, members: Dict[str, Any]):
    DATA_MEMBERS[data_type].update(members)


def get_cls_members(data_type: DataType) -> Dict[str, Any]:
    if data_type not in DATA_MEMBERS:  # pragma: no cover
        raise ValueError(f"{data_type} do not have any bound class member.")

    return DATA_MEMBERS[data_type]


def replace_warning_msg_on_no_session():
    session.warning_msg = (
        """No existing session found, creating a new local session now."""
    )


replace_warning_msg_on_no_session()



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, Union

import cloudpickle
import numpy as np
import pandas as pd

from ... import opcodes
from ...core import OutputType
from ...core.context import Context
from ...core.custom_log import redirect_custom_log
from ...serialization.serializables import AnyField, DictField, StringField
from ...utils import enter_current_session, get_func_token, quiet_stdio
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import build_df, build_empty_df, parse_index


class DataFrameApplymap(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = opcodes.APPLYMAP

    func = AnyField("func")
    na_action = StringField("na_action", default=None)
    func_token = StringField("func_token", default=None)
    kwds = DictField("kwds", default=None)

    def __init__(self, **kw):
        super().__init__(**kw)

    def _load_func(self):
        if isinstance(self.func, bytes):
            return cloudpickle.loads(self.func)
        else:
            return self.func

    @classmethod
    @redirect_custom_log
    @enter_current_session
    def execute(cls, ctx: Union[dict, Context], op: "DataFrameApplymap"):
        func = op._load_func()
        input_data = ctx[op.inputs[0].key]
        out = op.outputs[0]
        if len(input_data) == 0:
            ctx[out.key] = build_empty_df(out.dtypes)

        result = input_data.applymap(
            func,
            na_action=op.na_action,
            **op.kwds,
        )
        ctx[out.key] = result

    @classmethod
    def tile(cls, op: "DataFrameApplymap"):
        in_df = op.inputs[0]
        out_df = op.outputs[0]
        chunks = []
        for c in in_df.chunks:
            new_shape = c.shape

            new_index_value, new_columns_value = c.index_value, c.columns_value

            new_dtypes = out_df.dtypes

            new_op = op.copy().reset_key()
            new_op.tileable_op_key = op.key
            chunks.append(
                new_op.new_chunk(
                    [c],
                    shape=tuple(new_shape),
                    index=c.index,
                    dtypes=new_dtypes,
                    index_value=new_index_value,
                    columns_value=new_columns_value,
                )
            )
        new_nsplits = list(in_df.nsplits)
        new_op = op.copy()
        kw = out_df.params.copy()
        new_nsplits = tuple(new_nsplits)
        kw.update(dict(chunks=chunks, nsplits=new_nsplits))
        return new_op.new_tileables(op.inputs, **kw)

    def _infer_df_func_returns(self, df: pd.DataFrame):
        func = self._load_func()
        if isinstance(func, np.ufunc):
            output_type = OutputType.dataframe
            new_dtypes = None
            index_value = "inherit"
        else:
            output_type = new_dtypes = index_value = None

        try:
            empty_df = build_df(df, size=2)
            with np.errstate(all="ignore"), quiet_stdio():
                infer_df = empty_df.applymap(
                    func,
                    na_action=self.na_action,
                    **self.kwds,
                )
            if index_value is None:
                if infer_df.index is empty_df.index:
                    index_value = "inherit"
                else:  # pragma: no cover
                    index_value = parse_index(pd.RangeIndex(-1))

            output_type = output_type or OutputType.dataframe
            new_dtypes = new_dtypes or infer_df.dtypes
        except:  # noqa: E722  # nosec
            pass

        self.output_types = (
            [output_type] if not self.output_types else self.output_types
        )
        dtypes = new_dtypes
        return dtypes, index_value

    def __call__(self, df: pd.DataFrame, skip_infer: bool = False):
        self.func_token = get_func_token(self.func)
        if skip_infer:
            self.func = cloudpickle.dumps(self.func)
            return self.new_dataframe([df])

        # for backward compatibility
        dtypes, index_value = self._infer_df_func_returns(df)

        if index_value is None:
            index_value = parse_index(None, (df.key, df.index_value.key))
        for arg, desc in zip((self.output_types, dtypes), ("output_types", "dtypes")):
            if arg is None:
                raise TypeError(
                    f"Cannot determine {desc} by calculating with enumerate data, "
                    "please specify it as arguments"
                )

        if index_value == "inherit":
            index_value = df.index_value

        shape = df.shape

        self.func = cloudpickle.dumps(self.func)

        return self.new_dataframe(
            [df],
            shape=shape,
            dtypes=dtypes,
            index_value=index_value,
            columns_value=parse_index(dtypes.index, store_data=True),
        )


def df_applymap(
    df: pd.DataFrame,
    func: Callable,
    na_action: str = None,
    skip_infer: bool = False,
    **kwds,
):
    """
    Apply a function to a Dataframe elementwise.

    This method applies a function that accepts and returns a scalar
    to every element of a DataFrame.

    Parameters
    ----------
    func : callable
        Python function, returns a single value from a single value.

    na_action : {None, 'ignore'}, default None
        If 'ignore', propagate NaN values, without passing them to func.

    skip_infer: bool, default False
        Whether infer dtypes when dtypes or output_type is not specified.

    **kwds
        Additional keyword arguments to pass as keywords arguments to
        `func`.

    Returns
    -------
    DataFrame
        Transformed DataFrame.

    See Also
    --------
    DataFrame.apply : Apply a function along input axis of DataFrame.

    Examples
    --------
    >>> df = pd.DataFrame([[1, 2.12], [3.356, 4.567]])
    >>> df
            0      1
    0  1.000  2.120
    1  3.356  4.567

    >>> df.applymap(lambda x: len(str(x)))
        0  1
    0  3  4
    1  5  5

    Like Series.map, NA values can be ignored:

    >>> df_copy = df.copy()
    >>> df_copy.iloc[0, 0] = pd.NA
    >>> df_copy.applymap(lambda x: len(str(x)), na_action='ignore')
            0  1
    0  NaN  4
    1  5.0  5

    Note that a vectorized version of `func` often exists, which will
    be much faster. You could square each number elementwise.

    >>> df.applymap(lambda x: x**2)
                0          1
    0   1.000000   4.494400
    1  11.262736  20.857489

    But it's better to avoid applymap in that case.

    >>> df ** 2
                0          1
    0   1.000000   4.494400
    1  11.262736  20.857489
    """
    if na_action not in {"ignore", None}:
        raise ValueError(f"na_action must be 'ignore' or None. Got {repr(na_action)}")

    op = DataFrameApplymap(func=func, na_action=na_action, kwds=kwds)
    return op(df, skip_infer=skip_infer)



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
from functools import partial
from typing import Union
import binascii

import numpy as np
import pandas as pd

from .._mars import opcodes
from .._mars.core import recursive_tile
from .._mars.core.context import Context
from .._mars.core.entity import OutputType
from .._mars.core.operand import ObjectOperand, ObjectOperandMixin, OperandStage
from .._mars.dataframe.operands import DataFrameOperand, DataFrameOperandMixin
from .._mars.dataframe.utils import build_concatenated_rows_frame
from .._mars.serialization.serializables import AnyField
from .._mars.utils import CUnionFind as UnionFind
from ..core.adapter import from_mars, to_mars
from .utils import MERSENNE_PRIME, minhash_embed_func, optimal_param


class DataFrameUnionFind(ObjectOperand, ObjectOperandMixin):
    _output_type_ = OutputType.object
    union_find = AnyField("union_find", default=None)

    @classmethod
    def execute(cls, ctx: Union[dict, Context], op: "DataFrameUnionFind"):
        if op.stage == OperandStage.map:
            clusters = ctx[op.inputs[0].key]
            out = op.outputs[0]

            for cluster in clusters:
                if len(cluster) <= 1:
                    continue

                idx = min(cluster)
                for x in cluster:
                    op.union_find.union_(x, idx)

            ctx[out.key] = op.union_find
        elif op.stage == OperandStage.reduce:
            out = op.outputs[0]

            for i in range(len(op.inputs)):
                op.union_find.union_uf(ctx[op.inputs[i].key])
            ctx[out.key] = op.union_find


class DataFrameDedup(DataFrameOperand, DataFrameOperandMixin):
    _op_type = opcodes.DEDUP

    func = AnyField("func")

    @classmethod
    def execute(cls, ctx: Union[dict, Context], op: "DataFrameDedup"):
        input_data = ctx[op.inputs[0].key]
        uf = ctx[op.inputs[1].key]
        out = op.outputs[0]

        input_data["group_id"] = input_data["__dedup_id"].map(
            lambda x: binascii.hexlify(uf.find(x)).decode('utf-8')
        )

        ctx[out.key] = input_data.drop(columns="__dedup_id")

    @classmethod
    def tile(cls, op: "DataFrameDedup"):
        in_df = build_concatenated_rows_frame(op.inputs[0])
        out_df = op.outputs[0]

        def gen_id_column(df):
            from xoscar._utils import new_random_id

            df["__dedup_id"] = [new_random_id(16) for _ in range(len(df))]

            return df

        new_dtypes = in_df.dtypes.copy()
        new_dtypes["__dedup_id"] = "str"

        in_df_with_id = in_df.map_chunk(
            gen_id_column, output_type="dataframe", dtypes=new_dtypes
        )

        in_df_with_id = yield from recursive_tile(in_df_with_id)
        yield in_df_with_id.chunks

        embedded = in_df_with_id.apply(
            op.func,
            axis=1,
            output_type="dataframe",
            dtypes=pd.Series(["object", "bytes"], index=["__signatures", "__id"]),
        )

        clusters = (
            embedded.explode("__signatures")
            .groupby("__signatures", sort=False)["__id"]
            .apply(set)
        )
        tiled_clusters = yield from recursive_tile(clusters)

        # union find stage
        chunks = []
        for c in tiled_clusters.chunks:
            new_op = DataFrameUnionFind(union_find=UnionFind())
            new_op.stage = OperandStage.map
            chunks.append(
                new_op.new_chunk(
                    [c],
                    index=c.index,
                )
            )

        combine_size = 4
        while len(chunks) > combine_size:  # pragma: no cover
            new_chunks = []
            for i in range(0, len(chunks), combine_size):
                chks = chunks[i : i + combine_size]
                if len(chks) == 1:
                    chk = chks[0]
                else:
                    union_op = DataFrameUnionFind(union_find=UnionFind())
                    union_op.stage = OperandStage.reduce
                    for j, c in enumerate(chks):
                        c._index = (j, 0)
                    chk = union_op.new_chunk(chks)
                new_chunks.append(chk)
            chunks = new_chunks

        new_op = DataFrameUnionFind(union_find=UnionFind())
        new_op.stage = OperandStage.reduce
        union_chunk = new_op.new_chunk(chunks, index=(0,))
        union_chunk.is_broadcaster = True

        # dedup stage
        dedup_chunks = []
        for c in in_df_with_id.chunks:
            new_shape = c.shape

            new_op = op.copy().reset_key()

            dedup_chunks.append(
                new_op.new_chunk(
                    [c, union_chunk],
                    shape=(np.nan, new_shape[1] - 1),
                    index=c.index,
                    dtypes=out_df.dtypes,
                    index_value=c.index_value,
                    columns_value=out_df.columns_value,
                )
            )

        new_nsplits = tuple(chunk.shape[0] for chunk in dedup_chunks), (
            dedup_chunks[0].shape[1],
        )

        new_op = op.copy()
        kw = out_df.params.copy()
        kw.update(dict(chunks=dedup_chunks, nsplits=new_nsplits))

        return new_op.new_tileables(op.inputs, **kw)

    def __call__(self, df: pd.DataFrame):
        return self.new_dataframe([df])


def dedup(
    df: pd.DataFrame,
    col: str,
    method: str = "minhash",
    **kws,
) -> pd.DataFrame:
    """
    Applies deduplication on a DataFrame based on the chosen method.

    This function provides two methods for deduplication: exact matching and MinHash-based.
    The exact matching uses md5 hashing for deduplication, while the MinHash-based method
    utilizes MinHash and MinHashLSH for identifying and removing duplicates based on Jaccard similarity.
    For the MinHash-based method, it operates by generating hash values for a specified column
    of the DataFrame, computing similarity between these hash values, and then removing the rows
    that are determined to be duplicates according to a provided Jaccard similarity threshold.

    Parameters
    ----------
    df: pd.DataFrame,
        The DataFrame to deduplicate.

    col : str
        The column of the DataFrame on which to calculate hash values.

    method : str, default "minhash"
        The method for deduplication. Options include 'exact' and 'minhash'.

    Additional Parameters for MinHash method
    ----------------------------------------
    threshold : float, default 0.7
        The Jaccard similarity threshold to use in the MinHashLSH.

    num_perm : int, default 128
        The number of permutations to use in the MinHash.

    min_length : int, default 5
        The minimum number of tokens to use in the MinHash. Texts shorter than
        this value will be filtered out.

    ngrams : int, default 5
        The size of ngram to use in the MinHash.

    seed : int, default 42
        The seed for the random number generator.

    Returns
    -------
    DataFrame
        The DataFrame after applying the chosen deduplication method.

    Notes
    -----
    The 'exact' method performs deduplication by hashing each entry in the specified column with md5
    and removing duplicates.

    The 'minhash' method uses a combination of MinHash and MinHashLSH for efficient calculation of
    Jaccard similarity and identification of duplicates. This process involves hashing text to a
    finite set of integers (hash values), and then comparing these hash values to find duplicates.

    The optimal parameters for the number of bands `B` and rows `R` per band
    are automatically calculated based on the provided similarity threshold and
    number of permutations, to balance the trade-off between precision and recall.

    Examples
    --------
    >>> from xorbits.experimental import dedup
    >>> words = list("abcdefghijklmnopqrstuvwxyz")
    >>> df = pd.DataFrame(
    ...     {
    ...         "text": [
    ...             " ".join(["".join(np.random.choice(words, 5)) for i in range(50)])
    ...             for _ in np.arange(10)
    ...         ]
    ...         * 2,
    ...     }
    ... )
    >>> res = dedup(df, col="text", method="exact") # for 'exact' method
    >>> res = dedup(df, col="text", method="minhash", threshold=0.8, num_perm=128, min_length=5, ngrams=5, seed=42) # for 'minhash' method
    """

    if method not in ["exact", "minhash"]:
        raise ValueError("method must be one of 'exact' or 'minhash'")

    # Check if the DataFrame contains the text column
    if col not in df.dtypes.index:
        raise ValueError(f"{col} column not found in the DataFrame")

    if method == "exact":
        df = to_mars(df)

        df["__exact"] = df[col].apply(
            lambda x: hashlib.md5(x.encode("utf-8")).hexdigest()
        )

        df.drop_duplicates(subset=["__exact"], inplace=True).drop(
            columns=["__exact"], inplace=True
        )

        return from_mars(df)

    if method == "minhash":
        threshold = kws.pop("threshold", 0.7)
        num_perm = kws.pop("num_perm", 128)
        min_length = kws.pop("min_length", 5)
        ngrams = kws.pop("ngrams", 5)
        seed = kws.pop("seed", 42)

        # Check the threshold type and range
        if not isinstance(threshold, (float, int)) or not 0 <= threshold <= 1:
            raise ValueError(
                f"Expected 'threshold' to be a float between 0 and 1, got {threshold}"
            )

        # Check the num_perm, min_length, ngram and seed type and value
        for var, var_name in [
            (num_perm, "num_perm"),
            (min_length, "min_length"),
            (ngrams, "ngrams"),
            (seed, "seed"),
        ]:
            if not isinstance(var, int) or var <= 0:
                raise ValueError(
                    f"Expected '{var_name}' to be a positive integer, got {var}"
                )

        B, R = optimal_param(threshold, num_perm)

        HASH_RANGES = [(i * R, (i + 1) * R) for i in range(B)]

        RNG = np.random.RandomState(seed)

        PERMUTATIONS = np.array(
            [
                (
                    RNG.randint(1, MERSENNE_PRIME, dtype=np.uint64),
                    RNG.randint(0, MERSENNE_PRIME, dtype=np.uint64),
                )
                for _ in range(num_perm)
            ],
            dtype=np.uint64,
        ).T

        func = partial(
            minhash_embed_func,
            text=col,
            num_perm=num_perm,
            hashranges=HASH_RANGES,
            ngram_size=ngrams,
            min_length=min_length,
            permutations=PERMUTATIONS,
        )

        op = DataFrameDedup(func=func)

        return from_mars(op(to_mars(df)))



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numbers
from math import log, sqrt

import numpy as np
from scipy.special import gammaln
from sklearn.utils.extmath import fast_logdet
from sklearn.utils.validation import check_is_fitted

from ... import remote as mr
from ... import tensor as mt
from ...core import ExecutableTuple
from ...lib.sparse import issparse
from ...tensor.array_utils import get_array_module
from ...tensor.core import TENSOR_TYPE
from ...tensor.linalg import randomized_svd
from ...tensor.linalg.randomized_svd import svd_flip
from ...tensor.utils import check_random_state
from ..utils import check_array
from ._base import _BasePCA


def _assess_dimension(spectrum, rank, n_samples):
    """Compute the log-likelihood of a rank ``rank`` dataset.

    The dataset is assumed to be embedded in gaussian noise of shape(n,
    dimf) having spectrum ``spectrum``.

    Parameters
    ----------
    spectrum : array of shape (n_features)
        Data spectrum.
    rank : int
        Tested rank value. It should be strictly lower than n_features,
        otherwise the method isn't specified (division by zero in equation
        (31) from the paper).
    n_samples : int
        Number of samples.

    Returns
    -------
    ll : float,
        The log-likelihood

    Notes
    -----
    This implements the method of `Thomas P. Minka:
    Automatic Choice of Dimensionality for PCA. NIPS 2000: 598-604`
    """

    xp = get_array_module(spectrum, nosparse=True)

    n_features = spectrum.shape[0]
    if not 1 <= rank < n_features:  # pragma: no cover
        raise ValueError("the tested rank should be in [1, n_features - 1]")

    eps = 1e-15

    if spectrum[rank - 1] < eps:  # pragma: no cover
        # When the tested rank is associated with a small eigenvalue, there's
        # no point in computing the log-likelihood: it's going to be very
        # small and won't be the max anyway. Also, it can lead to numerical
        # issues below when computing pa, in particular in log((spectrum[i] -
        # spectrum[j]) because this will take the log of something very small.
        return -np.inf

    pu = -rank * log(2.0)
    for i in range(1, rank + 1):
        pu += (
            gammaln((n_features - i + 1) / 2.0)
            - log(np.pi) * (n_features - i + 1) / 2.0
        )

    pl = xp.sum(xp.log(spectrum[:rank]))
    pl = -pl * n_samples / 2.0

    v = max(eps, xp.sum(spectrum[rank:]) / (n_features - rank))
    pv = -xp.log(v) * n_samples * (n_features - rank) / 2.0

    m = n_features * rank - rank * (rank + 1.0) / 2.0
    pp = log(2.0 * np.pi) * (m + rank) / 2.0

    pa = 0.0
    spectrum_ = spectrum.copy()
    spectrum_[rank:n_features] = v
    for i in range(rank):
        for j in range(i + 1, len(spectrum)):
            pa += log(
                (spectrum[i] - spectrum[j]) * (1.0 / spectrum_[j] - 1.0 / spectrum_[i])
            ) + log(n_samples)

    ll = pu + pl + pv + pp - pa / 2.0 - rank * log(n_samples) / 2.0

    return ll


def _infer_dimension(spectrum, n_samples):
    """Infers the dimension of a dataset with a given spectrum.

    The returned value will be in [1, n_features - 1].
    """
    xp = get_array_module(spectrum, nosparse=True)

    ll = xp.empty_like(spectrum)
    ll[0] = -np.inf  # we don't want to return n_components = 0
    for rank in range(1, spectrum.shape[0]):
        ll[rank] = _assess_dimension(spectrum, rank, n_samples)
    return ll.argmax()


class PCA(_BasePCA):
    """Principal component analysis (PCA)

    Linear dimensionality reduction using Singular Value Decomposition of the
    data to project it to a lower dimensional space. The input data is centered
    but not scaled for each feature before applying the SVD.

    It uses the LAPACK implementation of the full SVD or a randomized truncated
    SVD by the method of Halko et al. 2009, depending on the shape of the input
    data and the number of components to extract.

    It can also use the scipy.sparse.linalg ARPACK implementation of the
    truncated SVD.

    Notice that this class does not support sparse input. See
    :class:`TruncatedSVD` for an alternative with sparse data.

    Read more in the :ref:`User Guide <PCA>`.

    Parameters
    ----------
    n_components : int, float, None or string
        Number of components to keep.
        if n_components is not set all components are kept::

            n_components == min(n_samples, n_features)

        If ``n_components == 'mle'`` and ``svd_solver == 'full'``, Minka's
        MLE is used to guess the dimension. Use of ``n_components == 'mle'``
        will interpret ``svd_solver == 'auto'`` as ``svd_solver == 'full'``.

        If ``0 < n_components < 1`` and ``svd_solver == 'full'``, select the
        number of components such that the amount of variance that needs to be
        explained is greater than the percentage specified by n_components.

        If ``svd_solver == 'arpack'``, the number of components must be
        strictly less than the minimum of n_features and n_samples.

        Hence, the None case results in::

            n_components == min(n_samples, n_features) - 1

    copy : bool (default True)
        If False, data passed to fit are overwritten and running
        fit(X).transform(X) will not yield the expected results,
        use fit_transform(X) instead.

    whiten : bool, optional (default False)
        When True (False by default) the `components_` vectors are multiplied
        by the square root of n_samples and then divided by the singular values
        to ensure uncorrelated outputs with unit component-wise variances.

        Whitening will remove some information from the transformed signal
        (the relative variance scales of the components) but can sometime
        improve the predictive accuracy of the downstream estimators by
        making their data respect some hard-wired assumptions.

    svd_solver : string {'auto', 'full', 'arpack', 'randomized'}
        auto :
            the solver is selected by a default policy based on `X.shape` and
            `n_components`: if the input data is larger than 500x500 and the
            number of components to extract is lower than 80% of the smallest
            dimension of the data, then the more efficient 'randomized'
            method is enabled. Otherwise the exact full SVD is computed and
            optionally truncated afterwards.
        full :
            run exact full SVD calling the standard LAPACK solver via
            `scipy.linalg.svd` and select the components by postprocessing
        arpack :
            run SVD truncated to n_components calling ARPACK solver via
            `scipy.sparse.linalg.svds`. It requires strictly
            0 < n_components < min(X.shape)
        randomized :
            run randomized SVD by the method of Halko et al.

    tol : float >= 0, optional (default .0)
        Tolerance for singular values computed by svd_solver == 'arpack'.

    iterated_power : int >= 0, or 'auto', (default 'auto')
        Number of iterations for the power method computed by
        svd_solver == 'randomized'.

    random_state : int, RandomState instance or None, optional (default None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`. Used when ``svd_solver`` == 'arpack' or 'randomized'.

    Attributes
    ----------
    components_ : tensor, shape (n_components, n_features)
        Principal axes in feature space, representing the directions of
        maximum variance in the data. The components are sorted by
        ``explained_variance_``.

    explained_variance_ : tensor, shape (n_components,)
        The amount of variance explained by each of the selected components.

        Equal to n_components largest eigenvalues
        of the covariance matrix of X.

    explained_variance_ratio_ : tensor, shape (n_components,)
        Percentage of variance explained by each of the selected components.

        If ``n_components`` is not set then all components are stored and the
        sum of the ratios is equal to 1.0.

    singular_values_ : tensor, shape (n_components,)
        The singular values corresponding to each of the selected components.
        The singular values are equal to the 2-norms of the ``n_components``
        variables in the lower-dimensional space.

    mean_ : tensor, shape (n_features,)
        Per-feature empirical mean, estimated from the training set.

        Equal to `X.mean(axis=0)`.

    n_components_ : int
        The estimated number of components. When n_components is set
        to 'mle' or a number between 0 and 1 (with svd_solver == 'full') this
        number is estimated from input data. Otherwise it equals the parameter
        n_components, or the lesser value of n_features and n_samples
        if n_components is None.

    noise_variance_ : float
        The estimated noise covariance following the Probabilistic PCA model
        from Tipping and Bishop 1999. See "Pattern Recognition and
        Machine Learning" by C. Bishop, 12.2.1 p. 574 or
        http://www.miketipping.com/papers/met-mppca.pdf. It is required to
        compute the estimated data covariance and score samples.

        Equal to the average of (min(n_features, n_samples) - n_components)
        smallest eigenvalues of the covariance matrix of X.

    References
    ----------
    For n_components == 'mle', this class uses the method of *Minka, T. P.
    "Automatic choice of dimensionality for PCA". In NIPS, pp. 598-604*

    Implements the probabilistic PCA model from:
    Tipping, M. E., and Bishop, C. M. (1999). "Probabilistic principal
    component analysis". Journal of the Royal Statistical Society:
    Series B (Statistical Methodology), 61(3), 611-622.
    via the score and score_samples methods.
    See http://www.miketipping.com/papers/met-mppca.pdf

    For svd_solver == 'arpack', refer to `scipy.sparse.linalg.svds`.

    For svd_solver == 'randomized', see:
    *Halko, N., Martinsson, P. G., and Tropp, J. A. (2011).
    "Finding structure with randomness: Probabilistic algorithms for
    constructing approximate matrix decompositions".
    SIAM review, 53(2), 217-288.* and also
    *Martinsson, P. G., Rokhlin, V., and Tygert, M. (2011).
    "A randomized algorithm for the decomposition of matrices".
    Applied and Computational Harmonic Analysis, 30(1), 47-68.*


    Examples
    --------
    >>> import mars.tensor as mt
    >>> from mars.learn.decomposition import PCA
    >>> X = mt.array([[-1, -1], [-2, -1], [-3, -2], [1, 1], [2, 1], [3, 2]])
    >>> pca = PCA(n_components=2)
    >>> pca.fit(X)  # doctest: +NORMALIZE_WHITESPACE
    PCA(copy=True, iterated_power='auto', n_components=2, random_state=None,
      svd_solver='auto', tol=0.0, whiten=False)
    >>> print(pca.explained_variance_ratio_)  # doctest: +ELLIPSIS
    [0.9924... 0.0075...]
    >>> print(pca.singular_values_)  # doctest: +ELLIPSIS
    [6.30061... 0.54980...]

    >>> pca = PCA(n_components=2, svd_solver='full')
    >>> pca.fit(X)                 # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
    PCA(copy=True, iterated_power='auto', n_components=2, random_state=None,
      svd_solver='full', tol=0.0, whiten=False)
    >>> print(pca.explained_variance_ratio_)  # doctest: +ELLIPSIS
    [0.9924... 0.00755...]
    >>> print(pca.singular_values_)  # doctest: +ELLIPSIS
    [6.30061... 0.54980...]

    See also
    --------
    KernelPCA
    SparsePCA
    TruncatedSVD
    IncrementalPCA
    """

    def __init__(
        self,
        n_components=None,
        copy=True,
        whiten=False,
        svd_solver="auto",
        tol=0.0,
        iterated_power="auto",
        random_state=None,
    ):
        self.n_components = n_components
        self.copy = copy
        self.whiten = whiten
        self.svd_solver = svd_solver
        self.tol = tol
        self.iterated_power = iterated_power
        self.random_state = random_state

    def fit(self, X, y=None, session=None, run_kwargs=None):
        """Fit the model with X.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where n_samples is the number of samples
            and n_features is the number of features.

        y : Ignored

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        self._fit(X, session=session, run=True, run_kwargs=run_kwargs)
        return self

    def fit_transform(self, X, y=None, session=None):
        """Fit the model with X and apply the dimensionality reduction on X.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where n_samples is the number of samples
            and n_features is the number of features.

        y : Ignored

        Returns
        -------
        X_new : array-like, shape (n_samples, n_components)

        """
        U, S, _ = self._fit(X, session=session, run=False)
        U = U[:, : self.n_components_]

        if self.whiten:
            # X_new = X * V / S * sqrt(n_samples) = U * sqrt(n_samples)
            U *= sqrt(X.shape[0] - 1)
        else:
            # X_new = X * V = U * S * V^T * V = U * S
            U *= S[: self.n_components_]

        self._run([U], session=session)
        return U

    def _run(self, result, session=None, run_kwargs=None):
        to_run_tensors = list(result)
        if isinstance(self.noise_variance_, TENSOR_TYPE):
            to_run_tensors.append(self.noise_variance_)
        to_run_tensors.append(self.components_)
        to_run_tensors.append(self.explained_variance_)
        to_run_tensors.append(self.explained_variance_ratio_)
        to_run_tensors.append(self.singular_values_)

        ExecutableTuple(to_run_tensors).execute(session=session, **(run_kwargs or {}))

    def _fit(self, X, session=None, run=True, run_kwargs=None):
        """Dispatch to the right submethod depending on the chosen solver."""

        # Raise an error for sparse input.
        # This is more informative than the generic one raised by check_array.
        if (hasattr(X, "issparse") and X.issparse()) or issparse(X):
            raise TypeError(
                "PCA does not support sparse input. See "
                "TruncatedSVD for a possible alternative."
            )

        X = check_array(
            X, dtype=[mt.float64, mt.float32], ensure_2d=True, copy=self.copy
        )

        # Handle n_components==None
        if self.n_components is None:
            if self.svd_solver != "arpack":
                n_components = min(X.shape)
            else:
                n_components = min(X.shape) - 1
        else:
            n_components = self.n_components

        # Handle svd_solver
        self._fit_svd_solver = self.svd_solver
        if self._fit_svd_solver == "auto":
            # Small problem or n_components == 'mle', just call full PCA
            if max(X.shape) <= 500 or n_components == "mle":
                self._fit_svd_solver = "full"
            elif n_components >= 1 and n_components < 0.8 * min(X.shape):
                self._fit_svd_solver = "randomized"
            # This is also the case of n_components in (0,1)
            else:
                self._fit_svd_solver = "full"

        # Call different fits for either full or truncated SVD
        if self._fit_svd_solver == "full":
            ret = self._fit_full(X, n_components, session=session)
        elif self._fit_svd_solver in ["arpack", "randomized"]:
            ret = self._fit_truncated(X, n_components, self._fit_svd_solver)
        else:
            raise ValueError(f"Unrecognized svd_solver='{self._fit_svd_solver}'")

        if run:
            self._run(ret, session=session, run_kwargs=run_kwargs)
        return ret

    def _fit_full(self, X, n_components, session=None, run_kwargs=None):
        """Fit the model by computing full SVD on X"""
        n_samples, n_features = X.shape

        if n_components == "mle":
            if n_samples < n_features:
                raise ValueError(
                    "n_components='mle' is only supported if n_samples >= n_features"
                )
        elif not 0 <= n_components <= min(n_samples, n_features):
            raise ValueError(
                "n_components=%r must be between 0 and "
                "min(n_samples, n_features)=%r with "
                "svd_solver='full'" % (n_components, min(n_samples, n_features))
            )
        elif n_components >= 1:
            if not isinstance(n_components, (numbers.Integral, np.integer)):
                raise ValueError(
                    "n_components=%r must be of type int "
                    "when greater than or equal to 1, "
                    "was of type=%r" % (n_components, type(n_components))
                )

        # Center data
        self.mean_ = mt.mean(X, axis=0)
        X -= self.mean_

        U, S, V = mt.linalg.svd(X)
        # flip eigenvectors' sign to enforce deterministic output
        U, V = svd_flip(U, V)

        components_ = V

        # Get variance explained by singular values
        explained_variance_ = (S**2) / (n_samples - 1)
        total_var = explained_variance_.sum()
        explained_variance_ratio_ = explained_variance_ / total_var
        singular_values_ = S.copy()  # Store the singular values.

        # Postprocess the number of components required
        if n_components == "mle":
            n_components = mr.spawn(
                _infer_dimension,
                args=(explained_variance_, n_samples),
                resolve_tileable_input=True,
            )
            ExecutableTuple([n_components, U, V]).execute(
                session=session, **(run_kwargs or dict())
            )
            n_components = n_components.fetch(session=session)
        elif 0 < n_components < 1.0:
            # number of components for which the cumulated explained
            # variance percentage is superior to the desired threshold
            # ratio_cumsum = stable_cumsum(explained_variance_ratio_)
            ratio_cumsum = explained_variance_ratio_.cumsum()
            n_components = (mt.searchsorted(ratio_cumsum, n_components) + 1).to_numpy(
                session=session, **(run_kwargs or dict())
            )

        # Compute noise covariance using Probabilistic PCA model
        # The sigma2 maximum likelihood (cf. eq. 12.46)
        if n_components < min(n_features, n_samples):
            self.noise_variance_ = explained_variance_[n_components:].mean()
        else:
            self.noise_variance_ = 0.0

        self.n_samples_, self.n_features_ = n_samples, n_features
        self.components_ = components_[:n_components]
        self.n_components_ = n_components
        self.explained_variance_ = explained_variance_[:n_components]
        self.explained_variance_ratio_ = explained_variance_ratio_[:n_components]
        self.singular_values_ = singular_values_[:n_components]

        return U, S, V

    def _fit_truncated(self, X, n_components, svd_solver):
        """Fit the model by computing truncated SVD (by ARPACK or randomized)
        on X
        """
        n_samples, n_features = X.shape

        if isinstance(n_components, str):
            raise ValueError(
                "n_components=%r cannot be a string "
                "with svd_solver='%s'" % (n_components, svd_solver)
            )
        elif not 1 <= n_components <= min(n_samples, n_features):
            raise ValueError(
                "n_components=%r must be between 1 and "
                "min(n_samples, n_features)=%r with "
                "svd_solver='%s'"
                % (n_components, min(n_samples, n_features), svd_solver)
            )
        elif not isinstance(n_components, (numbers.Integral, np.integer)):
            raise ValueError(
                "n_components=%r must be of type int "
                "when greater than or equal to 1, was of type=%r"
                % (n_components, type(n_components))
            )
        elif svd_solver == "arpack" and n_components == min(n_samples, n_features):
            raise ValueError(
                "n_components=%r must be strictly less than "
                "min(n_samples, n_features)=%r with "
                "svd_solver='%s'"
                % (n_components, min(n_samples, n_features), svd_solver)
            )

        random_state = check_random_state(self.random_state)

        # Center data
        self.mean_ = mt.mean(X, axis=0)
        X -= self.mean_

        if svd_solver == "arpack":
            # # random init solution, as ARPACK does it internally
            # v0 = random_state.uniform(-1, 1, size=min(X.shape))
            # U, S, V = svds(X, k=n_components, tol=self.tol, v0=v0)
            # # svds doesn't abide by scipy.linalg.svd/randomized_svd
            # # conventions, so reverse its outputs.
            # S = S[::-1]
            # # flip eigenvectors' sign to enforce deterministic output
            # U, V = svd_flip(U[:, ::-1], V[::-1])
            raise NotImplementedError("Does not support arpack svd_resolver")

        elif svd_solver == "randomized":
            # sign flipping is done inside
            U, S, V = randomized_svd(
                X,
                n_components=n_components,
                n_iter=self.iterated_power,
                flip_sign=True,
                random_state=random_state,
            )

        self.n_samples_, self.n_features_ = n_samples, n_features
        self.components_ = V
        self.n_components_ = n_components

        # Get variance explained by singular values
        self.explained_variance_ = (S**2) / (n_samples - 1)
        total_var = mt.var(X, ddof=1, axis=0)
        self.explained_variance_ratio_ = self.explained_variance_ / total_var.sum()
        self.singular_values_ = S.copy()  # Store the singular values.

        if self.n_components_ < min(n_features, n_samples):
            self.noise_variance_ = total_var.sum() - self.explained_variance_.sum()
            self.noise_variance_ /= min(n_features, n_samples) - n_components
        else:
            self.noise_variance_ = 0.0

        return U, S, V

    def _score_samples(self, X, session=None):
        check_is_fitted(self, "mean_")

        X = check_array(X)
        Xr = X - self.mean_
        n_features = X.shape[1]
        precision = self.get_precision().fetch(session=session)
        log_like = -0.5 * (Xr * (mt.dot(Xr, precision))).sum(axis=1)
        log_like -= 0.5 * (n_features * log(2.0 * mt.pi) - fast_logdet(precision))
        return log_like

    def score_samples(self, X, session=None):
        """Return the log-likelihood of each sample.

        See. "Pattern Recognition and Machine Learning"
        by C. Bishop, 12.2.1 p. 574
        or http://www.miketipping.com/papers/met-mppca.pdf

        Parameters
        ----------
        X : tensor, shape(n_samples, n_features)
            The data.

        Returns
        -------
        ll : tensor, shape (n_samples,)
            Log-likelihood of each sample under the current model
        """
        log_like = self._score_samples(X, session=session)
        log_like.execute(session=session)
        return log_like

    def score(self, X, y=None, session=None):
        """Return the average log-likelihood of all samples.

        See. "Pattern Recognition and Machine Learning"
        by C. Bishop, 12.2.1 p. 574
        or http://www.miketipping.com/papers/met-mppca.pdf

        Parameters
        ----------
        X : tensor, shape(n_samples, n_features)
            The data.

        y : Ignored

        Returns
        -------
        ll : float
            Average log-likelihood of the samples under the current model
        """
        ret = mt.mean(self._score_samples(X))
        ret.execute(session=session)
        return ret



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import logging

from ...core import TileStatus
from ...core.context import get_context
from ...serialization.serializables import KeyField
from ...typing import OperandType, TileableType
from ..core import DATAFRAME_TYPE, SERIES_TYPE
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import auto_merge_chunks


class DataFrameDeviceConversionBase(DataFrameOperand, DataFrameOperandMixin):
    _input = KeyField("input")

    @property
    def input(self):
        return self._input

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = inputs[0]

    def __call__(self, obj):
        if isinstance(obj, DATAFRAME_TYPE):
            return self.new_dataframe(
                [obj],
                shape=obj.shape,
                dtypes=obj.dtypes,
                index_value=obj.index_value,
                columns_value=obj.columns_value,
            )
        else:
            assert isinstance(obj, SERIES_TYPE)
            return self.new_series(
                [obj],
                shape=obj.shape,
                dtype=obj.dtype,
                index_value=obj.index_value,
                name=obj.name,
            )

    @classmethod
    def tile(cls, op):
        # Isolate ops on cpu from subsequent ops on gpu
        yield
        out_chunks = []
        for c in op.input.chunks:
            chunk_op = op.copy().reset_key()
            out_chunk = chunk_op.new_chunk([c], **c.params)
            out_chunks.append(out_chunk)

        new_op = op.copy().reset_key()
        out = op.outputs[0]
        return new_op.new_tileables(
            op.inputs, chunks=out_chunks, nsplits=op.inputs[0].nsplits, **out.params
        )


class DataFrameAutoMergeMixin(DataFrameOperandMixin):
    @classmethod
    def _get_auto_merge_options(cls, auto_merge: str) -> tuple[bool, bool]:
        if auto_merge == "both":
            return True, True
        elif auto_merge == "none":
            return False, False
        elif auto_merge == "before":
            return True, False
        else:
            assert auto_merge == "after"
            return False, True

    @classmethod
    def _merge_before(
        cls,
        op: OperandType,
        auto_merge_before: bool,
        auto_merge_threshold: int,
        left: TileableType,
        right: TileableType,
        logger: logging.Logger,
    ):
        ctx = get_context()

        if (
            auto_merge_before
            and len(left.chunks) + len(right.chunks) > auto_merge_threshold
        ):
            yield TileStatus([left, right] + left.chunks + right.chunks, progress=0.2)
            left_chunk_size = len(left.chunks)
            right_chunk_size = len(right.chunks)
            left = auto_merge_chunks(ctx, left)
            right = auto_merge_chunks(ctx, right)
            logger.info(
                "Auto merge before %s, left data shape: %s, chunk count: %s -> %s, "
                "right data shape: %s, chunk count: %s -> %s.",
                op,
                left.shape,
                left_chunk_size,
                len(left.chunks),
                right.shape,
                right_chunk_size,
                len(right.chunks),
            )
        else:
            logger.info(
                "Skip auto merge before %s, left data shape: %s, chunk count: %d, "
                "right data shape: %s, chunk count: %d.",
                op,
                left.shape,
                len(left.chunks),
                right.shape,
                len(right.chunks),
            )
        return [left, right]

    @classmethod
    def _merge_after(
        cls,
        op: OperandType,
        auto_merge_after: bool,
        auto_merge_threshold: int,
        ret: TileableType,
        logger: logging.Logger,
    ):
        if auto_merge_after and len(ret[0].chunks) > auto_merge_threshold:
            # if how=="inner", output data size will reduce greatly with high probability，
            # use auto_merge_chunks to combine small chunks.
            yield TileStatus(
                ret[0].chunks, progress=0.8
            )  # trigger execution for chunks
            merged = auto_merge_chunks(get_context(), ret[0])
            logger.info(
                "Auto merge after %s, data shape: %s, chunk count: %s -> %s.",
                op,
                merged.shape,
                len(ret[0].chunks),
                len(merged.chunks),
            )
            return [merged]
        else:
            logger.info(
                "Skip auto merge after %s, data shape: %s, chunk count: %d.",
                op,
                ret[0].shape,
                len(ret[0].chunks),
            )
            return ret



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import zipfile
from typing import Dict
from urllib.parse import urlparse

import numpy as np
import pandas as pd

try:
    import pyarrow as pa
    import pyarrow.parquet as pq

    arrow_dtype = pa.DataType
except ImportError:
    pa = None
    arrow_dtype = None

try:
    import fastparquet
except ImportError:
    fastparquet = None

import fsspec

from ... import opcodes as OperandDef
from ...config import options
from ...serialization.serializables import (
    AnyField,
    BoolField,
    DictField,
    Int32Field,
    Int64Field,
    ListField,
    StringField,
)
from ...utils import is_object_dtype, lazy_import
from ..operands import OutputType
from ..utils import PD_VERSION_GREATER_THAN_2_10, arrow_dtype_kwargs, parse_index
from .core import (
    ColumnPruneSupportedDataSourceMixin,
    IncrementalIndexDatasource,
    IncrementalIndexDataSourceMixin,
    merge_small_files,
)
from .utils import convert_to_abspath

PARQUET_MEMORY_SCALE = 10
PARQUET_MEMORY_SCALE_WITH_ARROW_DTYPE = 3
STRING_FIELD_OVERHEAD = 50
cudf = lazy_import("cudf")


def check_engine(engine):
    if engine == "auto":
        if pa is not None:
            return "pyarrow"
        elif fastparquet is not None:  # pragma: no cover
            return "fastparquet"
        else:  # pragma: no cover
            raise RuntimeError("Please install either pyarrow or fastparquet.")
    elif engine == "pyarrow":
        if pa is None:  # pragma: no cover
            raise RuntimeError("Please install pyarrow first.")
        return engine
    elif engine == "fastparquet":
        if fastparquet is None:  # pragma: no cover
            raise RuntimeError("Please install fastparquet first.")
        return engine
    else:  # pragma: no cover
        raise RuntimeError("Unsupported engine {} to read parquet.".format(engine))


def get_engine(engine):
    if engine == "pyarrow":
        return ArrowEngine()
    elif engine == "fastparquet":
        return FastpaquetEngine()
    else:  # pragma: no cover
        raise RuntimeError("Unsupported engine {}".format(engine))


class ParquetEngine:
    """Read parquet by arrow / fastparquet instead of pandas is to read the
    parquet file by group, please refer to `groups_as_chunks`."""

    def get_row_num(self, f):
        raise NotImplementedError

    def read_dtypes(self, f, **kwargs):
        raise NotImplementedError

    def read_to_pandas(
        self, f, columns=None, nrows=None, use_arrow_dtype=None, **kwargs
    ):
        raise NotImplementedError

    def read_group_to_pandas(
        self, f, group_index, columns=None, nrows=None, use_arrow_dtype=None, **kwargs
    ):
        raise NotImplementedError

    def read_partitioned_to_pandas(
        self,
        f,
        partitions: Dict,
        partition_keys: Dict,
        columns=None,
        nrows=None,
        use_arrow_dtype=None,
        **kwargs,
    ):
        raw_df = self.read_to_pandas(
            f, columns=columns, nrows=nrows, use_arrow_dtype=use_arrow_dtype, **kwargs
        )
        for col, value in partition_keys.items():
            dictionary = partitions[col]
            raw_df[col] = pd.Series(
                value,
                dtype=pd.CategoricalDtype(categories=dictionary.tolist()),
                index=raw_df.index,
            )
        return raw_df

    def read_partitioned_dtypes(
        self, fs: fsspec.AbstractFileSystem, directory, storage_options
    ):
        # As ParquetDataset will iterate all files,
        # here we just find one file to infer dtypes
        current_path = directory
        partition_cols = []
        while fs.isdir(current_path):
            _, dirs, files = next(fs.walk(current_path))
            dirs = [d for d in dirs if not d.startswith(".")]
            files = [f for f in files if not f.startswith(".")]
            if len(files) == 0:
                # directory as partition
                partition_cols.append(dirs[0].split("=", 1)[0])
                current_path = os.path.join(current_path, dirs[0])
            elif len(dirs) == 0:
                # parquet files in deepest directory
                current_path = os.path.join(current_path, files[0])
            else:  # pragma: no cover
                raise ValueError(
                    "Files and directories are mixed in an intermediate directory"
                )

        # current path is now a parquet file
        of = fsspec.open(current_path, storage_options=storage_options)
        with of as f:
            dtypes = self.read_dtypes(f)
        for partition in partition_cols:
            dtypes[partition] = pd.CategoricalDtype()
        return dtypes


def _parse_prefix(path):
    path_prefix = ""
    if isinstance(path, str):
        parsed_path = urlparse(path)
        if parsed_path.scheme:
            path_prefix = f"{parsed_path.scheme}://{parsed_path.netloc}"
    return path_prefix


class ArrowEngine(ParquetEngine):
    def get_row_num(self, f):
        file = pq.ParquetFile(f)
        return file.metadata.num_rows

    def read_dtypes(self, f, **kwargs):
        types_mapper = kwargs.pop("types_mapper", None)
        file = pq.ParquetFile(f)
        return (
            file.schema_arrow.empty_table().to_pandas(types_mapper=types_mapper).dtypes
        )

    @classmethod
    def _table_to_pandas(cls, t, nrows=None, use_arrow_dtype=None):
        if nrows is not None:
            t = t.slice(0, nrows)
        if use_arrow_dtype:
            df = t.to_pandas(types_mapper=pd.ArrowDtype)
        else:
            df = t.to_pandas()
        return df

    def read_to_pandas(
        self, f, columns=None, nrows=None, use_arrow_dtype=None, **kwargs
    ):
        file = pq.ParquetFile(f)
        t = file.read(columns=columns, **kwargs)
        return self._table_to_pandas(t, nrows=nrows, use_arrow_dtype=use_arrow_dtype)

    def read_group_to_pandas(
        self, f, group_index, columns=None, nrows=None, use_arrow_dtype=None, **kwargs
    ):
        file = pq.ParquetFile(f)
        t = file.read_row_group(group_index, columns=columns, **kwargs)
        return self._table_to_pandas(t, nrows=nrows, use_arrow_dtype=use_arrow_dtype)


class FastpaquetEngine(ParquetEngine):
    def get_row_num(self, f):
        file = fastparquet.ParquetFile(f)
        return file.count()

    def read_dtypes(self, f, **kwargs):
        file = fastparquet.ParquetFile(f)
        dtypes_dict = file._dtypes()
        return pd.Series(dict((c, dtypes_dict[c]) for c in file.columns))

    def read_to_pandas(
        self, f, columns=None, nrows=None, use_arrow_dtype=None, **kwargs
    ):
        file = fastparquet.ParquetFile(f)
        df = file.to_pandas(columns, **kwargs)
        if nrows is not None:
            df = df.head(nrows)
        return df


class CudfEngine:
    @classmethod
    def read_to_cudf(cls, file, columns: list = None, nrows: int = None, **kwargs):
        df = cudf.read_parquet(file, columns=columns, **kwargs)
        if nrows is not None:
            df = df.head(nrows)
        return df

    def read_group_to_cudf(
        self, file, group_index: int, columns: list = None, nrows: int = None, **kwargs
    ):
        return self.read_to_cudf(
            file, columns=columns, nrows=nrows, row_groups=group_index, **kwargs
        )

    @classmethod
    def read_partitioned_to_cudf(
        cls,
        file,
        partitions: Dict,
        partition_keys: Dict,
        columns=None,
        nrows=None,
        **kwargs,
    ):
        # cudf will read entire partitions even if only one partition provided,
        # so we just read with pyarrow and convert to cudf DataFrame
        file = pq.ParquetFile(file)
        t = file.read(columns=columns, **kwargs)
        t = t.slice(0, nrows) if nrows is not None else t
        t = pa.table(t.columns, names=t.column_names)
        raw_df = cudf.DataFrame.from_arrow(t)
        for col, value in partition_keys.items():
            dictionary = partitions[col].tolist()
            codes = cudf.core.column.as_column(
                dictionary.index(value), length=len(raw_df)
            )
            raw_df[col] = cudf.core.column.build_categorical_column(
                categories=dictionary,
                codes=codes,
                size=codes.size,
                offset=codes.offset,
                ordered=False,
            )
        return raw_df


class DataFrameReadParquet(
    IncrementalIndexDatasource,
    ColumnPruneSupportedDataSourceMixin,
    IncrementalIndexDataSourceMixin,
):
    _op_type_ = OperandDef.READ_PARQUET

    path = AnyField("path")
    chunk_path = AnyField("chunk_path")
    engine = StringField("engine")
    columns = ListField("columns")
    use_arrow_dtype = BoolField("use_arrow_dtype")
    groups_as_chunks = BoolField("groups_as_chunks")
    group_index = Int32Field("group_index")
    read_kwargs = DictField("read_kwargs")
    incremental_index = BoolField("incremental_index")
    storage_options = DictField("storage_options")
    is_partitioned = BoolField("is_partitioned")
    merge_small_files = BoolField("merge_small_files")
    merge_small_file_options = DictField("merge_small_file_options")
    is_http_url = BoolField("is_http_url", None)
    # for chunk
    partitions = DictField("partitions", default=None)
    partition_keys = DictField("partition_keys", default=None)
    num_group_rows = Int64Field("num_group_rows", default=None)
    # as read meta may be too time-consuming when number of files is large,
    # thus we only read first file to get row number and raw file size
    first_chunk_row_num = Int64Field("first_chunk_row_num")
    first_chunk_raw_bytes = Int64Field("first_chunk_raw_bytes")

    def get_columns(self):
        return self.columns

    def set_pruned_columns(self, columns, *, keep_order=None):
        self.columns = columns

    @classmethod
    def _tile_partitioned(cls, op: "DataFrameReadParquet"):
        out_df = op.outputs[0]
        shape = (np.nan, out_df.shape[1])
        dtypes = out_df.dtypes
        dataset = pq.ParquetDataset(op.path, use_legacy_dataset=False)

        path_prefix = _parse_prefix(op.path)

        chunk_index = 0
        out_chunks = []
        first_chunk_row_num, first_chunk_raw_bytes = None, None
        for i, fragment in enumerate(dataset.fragments):
            chunk_op = op.copy().reset_key()
            chunk_op.path = chunk_path = path_prefix + fragment.path
            relpath = os.path.relpath(chunk_path, op.path)
            partition_keys = dict(
                tuple(s.split("=")) for s in relpath.split(os.sep)[:-1]
            )
            chunk_op.partition_keys = partition_keys
            chunk_op.partitions = dict(
                zip(
                    dataset.partitioning.schema.names, dataset.partitioning.dictionaries
                )
            )
            if i == 0:
                first_row_group = fragment.row_groups[0]
                first_chunk_raw_bytes = first_row_group.total_byte_size
                first_chunk_row_num = first_row_group.num_rows
            chunk_op.first_chunk_row_num = first_chunk_row_num
            chunk_op.first_chunk_raw_bytes = first_chunk_raw_bytes
            new_chunk = chunk_op.new_chunk(
                None,
                shape=shape,
                index=(chunk_index, 0),
                index_value=out_df.index_value,
                columns_value=out_df.columns_value,
                dtypes=dtypes,
            )
            out_chunks.append(new_chunk)
            chunk_index += 1

        new_op = op.copy()
        nsplits = ((np.nan,) * len(out_chunks), (out_df.shape[1],))
        return new_op.new_dataframes(
            None,
            out_df.shape,
            dtypes=dtypes,
            index_value=out_df.index_value,
            columns_value=out_df.columns_value,
            chunks=out_chunks,
            nsplits=nsplits,
        )

    @classmethod
    def _tile_no_partitioned(cls, op: "DataFrameReadParquet"):
        chunk_index = 0
        out_chunks = []
        out_df = op.outputs[0]

        dtypes = out_df.dtypes
        shape = (np.nan, out_df.shape[1])
        z = None
        fs, _, _ = fsspec.get_fs_token_paths(
            op.path, storage_options=op.storage_options
        )
        if isinstance(op.path, (tuple, list)):
            paths = op.path
        elif fs.isdir(op.path):
            paths = fs.ls(op.path)
            paths = sorted(paths)
            if not isinstance(fs, fsspec.implementations.local.LocalFileSystem):
                parsed_path = urlparse(op.path)
                paths = [f"{parsed_path.scheme}://{path}" for path in paths]
        elif isinstance(op.path, str) and op.path.endswith(".zip"):
            file = fs.open(op.path, storage_options=op.storage_options)
            z = zipfile.ZipFile(file)
            paths = z.namelist()
            paths = [
                path
                for path in paths
                if path.endswith(".parquet") and not path.startswith("__MACOSX")
            ]
        else:
            paths = fs.glob(op.path, storage_options=op.storage_options)
            if not isinstance(fs, fsspec.implementations.local.LocalFileSystem):
                parsed_path = urlparse(op.path)
                paths = [f"{parsed_path.scheme}://{path}" for path in paths]
        first_chunk_row_num, first_chunk_raw_bytes = None, None
        for i, pth in enumerate(paths):
            if i == 0:
                if z is not None:
                    with z.open(pth) as f:
                        first_chunk_row_num = get_engine(op.engine).get_row_num(f)
                        first_chunk_raw_bytes = sys.getsizeof(f)
                else:
                    of = fsspec.open(pth)
                    with of as f:
                        first_chunk_row_num = get_engine(op.engine).get_row_num(f)
                    first_chunk_raw_bytes = fsspec.get_fs_token_paths(
                        pth, storage_options=op.storage_options
                    )[0].size(pth)

            if op.groups_as_chunks:
                if z is not None:
                    with z.open(pth) as f:
                        num_row_groups = pq.ParquetFile(f).num_row_groups
                else:
                    num_row_groups = pq.ParquetFile(pth).num_row_groups
                for group_idx in range(num_row_groups):
                    chunk_op = op.copy().reset_key()
                    if z is not None:
                        chunk_op.path = op.path
                        chunk_op.chunk_path = pth
                    else:
                        chunk_op.path = pth
                    chunk_op.group_index = group_idx
                    chunk_op.first_chunk_row_num = first_chunk_row_num
                    chunk_op.first_chunk_raw_bytes = first_chunk_raw_bytes
                    chunk_op.num_group_rows = num_row_groups
                    new_chunk = chunk_op.new_chunk(
                        None,
                        shape=shape,
                        index=(chunk_index, 0),
                        index_value=out_df.index_value,
                        columns_value=out_df.columns_value,
                        dtypes=dtypes,
                    )
                    out_chunks.append(new_chunk)
                    chunk_index += 1
            else:
                chunk_op = op.copy().reset_key()
                if z is not None:
                    chunk_op.path = op.path
                    chunk_op.chunk_path = pth
                else:
                    chunk_op.path = pth
                chunk_op.first_chunk_row_num = first_chunk_row_num
                chunk_op.first_chunk_raw_bytes = first_chunk_raw_bytes
                new_chunk = chunk_op.new_chunk(
                    None,
                    shape=shape,
                    index=(chunk_index, 0),
                    index_value=out_df.index_value,
                    columns_value=out_df.columns_value,
                    dtypes=dtypes,
                )
                out_chunks.append(new_chunk)
                chunk_index += 1
        if z is not None:
            z.close()
        new_op = op.copy()
        nsplits = ((np.nan,) * len(out_chunks), (out_df.shape[1],))
        return new_op.new_dataframes(
            None,
            out_df.shape,
            dtypes=dtypes,
            index_value=out_df.index_value,
            columns_value=out_df.columns_value,
            chunks=out_chunks,
            nsplits=nsplits,
        )

    @classmethod
    def _tile_http_url(cls, op: "DataFrameReadParquet"):
        out_chunks = []
        out_df = op.outputs[0]
        z = None
        if op.path[0].endswith(".zip"):
            fs, _, _ = fsspec.get_fs_token_paths(op.path[0])
            zip_filename = fs.open(op.path[0])
            z = zipfile.ZipFile(zip_filename)
            paths = z.namelist()
            paths = [
                path
                for path in paths
                if path.endswith(".parquet") and not path.startswith("__MACOSX")
            ]
        else:
            paths = op.path
        for i, url in enumerate(paths):
            chunk_op = op.copy().reset_key()
            if z is not None:
                chunk_op.path = op.path[0]
                chunk_op.chunk_path = url
            else:
                chunk_op.path = url
            out_chunks.append(
                chunk_op.new_chunk(None, index=(i, 0), shape=(np.nan, np.nan))
            )
        if z is not None:
            z.close()
        new_op = op.copy()
        nsplits = ((np.nan,) * len(out_chunks), (np.nan,))
        return new_op.new_dataframes(
            None,
            out_df.shape,
            dtypes=out_df.dtypes,
            index_value=out_df.index_value,
            columns_value=out_df.columns_value,
            chunks=out_chunks,
            nsplits=nsplits,
        )

    @classmethod
    def _tile(cls, op: "DataFrameReadParquet"):
        if op.is_http_url:
            tiled = cls._tile_http_url(op)
        elif op.is_partitioned:
            tiled = cls._tile_partitioned(op)
        else:
            tiled = cls._tile_no_partitioned(op)
        if op.merge_small_files:
            tiled = [
                merge_small_files(tiled[0], **(op.merge_small_file_options or dict()))
            ]
        return tiled

    @classmethod
    def _execute_partitioned(cls, ctx, op: "DataFrameReadParquet"):
        out = op.outputs[0]
        engine = get_engine(op.engine)
        of = fsspec.open(op.path, storage_options=op.storage_options)
        with of as f:
            ctx[out.key] = engine.read_partitioned_to_pandas(
                f,
                op.partitions,
                op.partition_keys,
                columns=op.columns,
                nrows=op.nrows,
                use_arrow_dtype=op.use_arrow_dtype,
                **op.read_kwargs or dict(),
            )

    @classmethod
    def _pandas_read_parquet(cls, ctx: dict, op: "DataFrameReadParquet"):
        out = op.outputs[0]
        path = op.path
        z = None
        if op.is_http_url:
            if op.path.endswith(".zip"):
                fs, _, _ = fsspec.get_fs_token_paths(op.path)
                zip_filename = fs.open(op.path)
                z = zipfile.ZipFile(zip_filename)
                f = z.open(op.chunk_path)
            else:
                f = op.path
            read_kwargs = op.read_kwargs or dict()
            if op.use_arrow_dtype:
                read_kwargs.update(arrow_dtype_kwargs())
            r = pd.read_parquet(
                f,
                columns=op.columns,
                engine=op.engine,
                **read_kwargs,
            )
            if z is not None:
                z.close()
                f.close()
            ctx[out.key] = r
            return
        if op.partitions is not None:
            return cls._execute_partitioned(ctx, op)
        engine = get_engine(op.engine)
        z = None
        fs = fsspec.get_fs_token_paths(path, storage_options=op.storage_options)[0]
        if op.path.endswith(".zip"):
            file = fs.open(op.path, storage_options=op.storage_options)
            z = zipfile.ZipFile(file)
            f = z.open(op.chunk_path)
        else:
            f = fs.open(path, storage_options=op.storage_options)
        use_arrow_dtype = op.use_arrow_dtype
        if op.groups_as_chunks:
            df = engine.read_group_to_pandas(
                f,
                op.group_index,
                columns=op.columns,
                nrows=op.nrows,
                use_arrow_dtype=use_arrow_dtype,
                **op.read_kwargs or dict(),
            )
        else:
            df = engine.read_to_pandas(
                f,
                columns=op.columns,
                nrows=op.nrows,
                use_arrow_dtype=use_arrow_dtype,
                **op.read_kwargs or dict(),
            )
        ctx[out.key] = df
        if z is not None:
            z.close()
        f.close()

    @classmethod
    def _cudf_read_parquet(cls, ctx: dict, op: "DataFrameReadParquet"):
        out = op.outputs[0]
        path = op.path
        z = None
        fs = fsspec.get_fs_token_paths(path, storage_options=op.storage_options)[0]
        if path.endswith(".zip"):
            z = zipfile.ZipFile(path)
        engine = CudfEngine()
        if os.path.exists(path) and z is None:
            file = op.path
            close = lambda: None
        else:  # pragma: no cover
            if z is not None:
                file = z.open(op.chunk_path)
            else:
                file = fs.open(path, storage_options=op.storage_options)
            close = file.close
        try:
            if op.partitions is not None:
                ctx[out.key] = engine.read_partitioned_to_cudf(
                    file,
                    op.partitions,
                    op.partition_keys,
                    columns=op.columns,
                    nrows=op.nrows,
                    **op.read_kwargs or dict(),
                )
            else:
                if op.groups_as_chunks:
                    df = engine.read_group_to_cudf(
                        file,
                        op.group_index,
                        columns=op.columns,
                        nrows=op.nrows,
                        **op.read_kwargs or dict(),
                    )
                else:
                    df = engine.read_to_cudf(
                        file,
                        columns=op.columns,
                        nrows=op.nrows,
                        **op.read_kwargs or dict(),
                    )
                ctx[out.key] = df
        finally:
            close()

    @classmethod
    def execute(cls, ctx, op: "DataFrameReadParquet"):
        if not op.gpu:
            cls._pandas_read_parquet(ctx, op)
        else:
            cls._cudf_read_parquet(ctx, op)

    @classmethod
    def estimate_size(cls, ctx, op: "DataFrameReadParquet"):
        if op.is_http_url:
            return super().estimate_size(ctx, op)
        first_chunk_row_num = op.first_chunk_row_num
        first_chunk_raw_bytes = op.first_chunk_raw_bytes
        if isinstance(op.path, str) and op.path.endswith(".zip"):
            with fsspec.open(op.path, storage_options=op.storage_options) as zip_file:
                with zipfile.ZipFile(zip_file) as z:
                    with z.open(op.chunk_path) as f:
                        raw_bytes = sys.getsizeof(f)
        else:
            raw_bytes = fsspec.get_fs_token_paths(
                op.path, storage_options=op.storage_options
            )[0].size(op.path)
        if op.num_group_rows:
            raw_bytes = (
                np.ceil(np.divide(raw_bytes, op.num_group_rows)).astype(np.int64).item()
            )

        estimated_row_num = (
            np.ceil(first_chunk_row_num * (raw_bytes / first_chunk_raw_bytes))
            .astype(np.int64)
            .item()
        )
        if op.columns is not None:
            of = fsspec.open(op.path, storage_options=op.storage_options)
            with of as f:
                all_columns = list(get_engine(op.engine).read_dtypes(f).index)
        else:
            all_columns = list(op.outputs[0].dtypes.index)
        columns = op.columns if op.columns else all_columns
        if op.use_arrow_dtype:
            scale = op.memory_scale or PARQUET_MEMORY_SCALE_WITH_ARROW_DTYPE
        else:
            scale = op.memory_scale or PARQUET_MEMORY_SCALE
        phy_size = raw_bytes * scale * len(columns) / len(all_columns)
        n_strings = len(
            [
                dt
                for col, dt in op.outputs[0].dtypes.items()
                if col in columns and is_object_dtype(dt)
            ]
        )
        if op.use_arrow_dtype:
            pd_size = phy_size
        else:
            pd_size = phy_size + n_strings * estimated_row_num * STRING_FIELD_OVERHEAD
        ctx[op.outputs[0].key] = (pd_size, pd_size)

    def __call__(self, index_value=None, columns_value=None, dtypes=None):
        self._output_types = [OutputType.dataframe]
        if dtypes is not None:
            shape = (np.nan, len(dtypes))
        else:
            shape = (np.nan, np.nan)
        return self.new_dataframe(
            None,
            shape,
            dtypes=dtypes,
            index_value=index_value,
            columns_value=columns_value,
        )


def read_parquet(
    path,
    engine: str = "auto",
    columns: list = None,
    groups_as_chunks: bool = False,
    use_arrow_dtype: bool = None,
    incremental_index: bool = False,
    storage_options: dict = None,
    memory_scale: int = None,
    merge_small_files: bool = True,
    merge_small_file_options: dict = None,
    gpu: bool = None,
    **kwargs,
):
    """
    Load a parquet object from the file path, returning a DataFrame.

    Parameters
    ----------
    path : str, path object or file-like object
        Any valid string path is acceptable. The string could be a URL.
        For file URLs, a host is expected. A local file could be:
        ``file://localhost/path/to/table.parquet``.
        A file URL can also be a path to a directory that contains multiple
        partitioned parquet files. Both pyarrow and fastparquet support
        paths to directories as well as file URLs. A directory path could be:
        ``file://localhost/path/to/tables``.
        By file-like object, we refer to objects with a ``read()`` method,
        such as a file handler (e.g. via builtin ``open`` function)
        or ``StringIO``.
    engine : {'auto', 'pyarrow', 'fastparquet'}, default 'auto'
        Parquet library to use. The default behavior is to try 'pyarrow',
        falling back to 'fastparquet' if 'pyarrow' is unavailable.
    columns : list, default=None
        If not None, only these columns will be read from the file.
    groups_as_chunks : bool, default False
        if True, each row group correspond to a chunk.
        if False, each file correspond to a chunk.
        Only available for 'pyarrow' engine.
    incremental_index: bool, default False
        If index_col not specified, ensure range index incremental,
        gain a slightly better performance if setting False.
    use_arrow_dtype: bool, default None
        If True, use arrow dtype to store columns. Default enabled if pandas >= 2.1
    storage_options: dict, optional
        Options for storage connection.
    memory_scale: int, optional
        Scale that real memory occupation divided with raw file size.
    merge_small_files: bool, default True
        Merge small files whose size is small.
    merge_small_file_options: dict
        Options for merging small files
    **kwargs
        Any additional kwargs are passed to the engine.

    Returns
    -------
    Mars DataFrame
    """

    engine_type = check_engine(engine)
    engine = get_engine(engine_type)

    # We enable arrow dtype by default if pandas >= 2.1
    if use_arrow_dtype is None and engine_type == "pyarrow":
        use_arrow_dtype = PD_VERSION_GREATER_THAN_2_10

    single_path = path[0] if isinstance(path, list) else path
    is_partitioned = False
    if isinstance(single_path, str) and (
        single_path.startswith("http://")
        or single_path.startswith("https://")
        or single_path.startswith("ftp://")
        or single_path.startswith("sftp://")
    ):
        urls = path if isinstance(path, (list, tuple)) else [path]
        op = DataFrameReadParquet(
            path=urls,
            engine=engine_type,
            columns=columns,
            groups_as_chunks=groups_as_chunks,
            use_arrow_dtype=use_arrow_dtype,
            read_kwargs=kwargs,
            incremental_index=incremental_index,
            storage_options=storage_options,
            memory_scale=memory_scale,
            merge_small_files=merge_small_files,
            merge_small_file_options=merge_small_file_options,
            is_http_url=True,
            gpu=gpu,
        )
        return op()
    fs, _, _ = fsspec.get_fs_token_paths(single_path, storage_options=storage_options)
    if use_arrow_dtype is None:
        use_arrow_dtype = options.dataframe.use_arrow_dtype
    if use_arrow_dtype and engine_type != "pyarrow":
        raise ValueError(
            f"The 'use_arrow_dtype' argument is not supported for the {engine_type} engine"
        )
    # We enable arrow dtype by default if pandas >= 2.1
    if use_arrow_dtype is None:
        use_arrow_dtype = PD_VERSION_GREATER_THAN_2_10

    types_mapper = pd.ArrowDtype if use_arrow_dtype else None

    if fs.isdir(single_path):
        paths = fs.ls(path)
        if all(fs.isdir(p) for p in paths):
            # If all are directories, it is read as a partitioned dataset.
            dtypes = engine.read_partitioned_dtypes(fs, path, storage_options)
            is_partitioned = True
        else:
            with fs.open(paths[0], mode="rb") as f:
                dtypes = engine.read_dtypes(f, types_mapper=types_mapper)
    elif isinstance(path, str) and path.endswith(".zip"):
        with fsspec.open(path, "rb") as file:
            with zipfile.ZipFile(file) as z:
                with z.open(z.namelist()[0]) as f:
                    dtypes = engine.read_dtypes(f, types_mapper=types_mapper)
    else:
        if not isinstance(path, list):
            file_path = fs.glob(path, storage_options=storage_options)[0]
        else:
            file_path = path[0]
        with fs.open(file_path, storage_options=storage_options) as f:
            dtypes = engine.read_dtypes(f, types_mapper=types_mapper)
    if columns:
        dtypes = dtypes[columns]

    index_value = parse_index(pd.RangeIndex(-1))
    columns_value = parse_index(dtypes.index, store_data=True)

    # convert path to abs_path
    abs_path = convert_to_abspath(path, storage_options)
    op = DataFrameReadParquet(
        path=abs_path,
        engine=engine_type,
        columns=columns,
        groups_as_chunks=groups_as_chunks,
        use_arrow_dtype=use_arrow_dtype,
        read_kwargs=kwargs,
        incremental_index=incremental_index,
        storage_options=storage_options,
        is_partitioned=is_partitioned,
        memory_scale=memory_scale,
        merge_small_files=merge_small_files,
        merge_small_file_options=merge_small_file_options,
        gpu=gpu,
    )
    return op(index_value=index_value, columns_value=columns_value, dtypes=dtypes)



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .beta import TensorRandBeta, beta
from .binomial import TensorBinomial, binomial
from .bytes import bytes
from .chisquare import TensorChisquareDist, chisquare
from .choice import TensorChoice, choice
from .core import RandomState, RandomStateField, _random_state
from .dirichlet import TensorDirichlet, dirichlet
from .exponential import TensorExponential, exponential
from .f import TensorF, f
from .gamma import TensorRandGamma, gamma
from .geometric import TensorGeometric, geometric
from .gumbel import TensorGumbel, gumbel
from .hypergeometric import TensorHypergeometric, hypergeometric
from .laplace import TensorLaplace, laplace
from .logistic import TensorLogistic, logistic
from .lognormal import TensorLognormal, lognormal
from .logseries import TensorLogseries, logseries
from .multinomial import TensorMultinomial, multinomial
from .multivariate_normal import TensorMultivariateNormal, multivariate_normal
from .negative_binomial import TensorNegativeBinomial, negative_binomial
from .noncentral_chisquare import TensorNoncentralChisquare, noncentral_chisquare
from .noncentral_f import TensorNoncentralF, noncentral_f
from .normal import TensorNormal, normal
from .pareto import TensorPareto, pareto
from .permutation import TensorPermutation, permutation
from .poisson import TensorPoisson, poisson
from .power import TensorRandomPower, power
from .rand import TensorRand, rand
from .randint import TensorRandint, randint
from .randn import TensorRandn, randn
from .random_integers import TensorRandomIntegers, random_integers
from .random_sample import TensorRandomSample, random_sample
from .rayleigh import TensorRayleigh, rayleigh
from .shuffle import shuffle
from .standard_cauchy import TensorStandardCauchy, standard_cauchy
from .standard_exponential import TensorStandardExponential, standard_exponential
from .standard_gamma import TensorStandardGamma, standard_gamma
from .standard_normal import TensorStandardNormal, standard_normal
from .standard_t import TensorStandardT, standard_t
from .triangular import TensorTriangular, triangular
from .uniform import TensorUniform, uniform
from .vonmises import TensorVonmises, vonmises
from .wald import TensorWald, wald
from .weibull import TensorWeibull, weibull
from .zipf import TensorZipf, zipf


def _install():
    setattr(RandomState, "rand", rand)
    setattr(RandomState, "randn", randn)
    setattr(RandomState, "randint", randint)
    setattr(RandomState, "random_integers", random_integers)
    setattr(RandomState, "random_sample", random_sample)
    setattr(RandomState, "ranf", random_sample)
    setattr(RandomState, "random", random_sample)
    setattr(RandomState, "sample", random_sample)
    setattr(RandomState, "choice", choice)
    setattr(RandomState, "bytes", bytes)
    setattr(RandomState, "beta", beta)
    setattr(RandomState, "binomial", binomial)
    setattr(RandomState, "chisquare", chisquare)
    setattr(RandomState, "dirichlet", dirichlet)
    setattr(RandomState, "exponential", exponential)
    setattr(RandomState, "f", f)
    setattr(RandomState, "gamma", gamma)
    setattr(RandomState, "geometric", geometric)
    setattr(RandomState, "gumbel", gumbel)
    setattr(RandomState, "hypergeometric", hypergeometric)
    setattr(RandomState, "laplace", laplace)
    setattr(RandomState, "logistic", logistic)
    setattr(RandomState, "lognormal", lognormal)
    setattr(RandomState, "logseries", logseries)
    setattr(RandomState, "multinomial", multinomial)
    setattr(RandomState, "multivariate_normal", multivariate_normal)
    setattr(RandomState, "negative_binomial", negative_binomial)
    setattr(RandomState, "noncentral_chisquare", noncentral_chisquare)
    setattr(RandomState, "noncentral_f", noncentral_f)
    setattr(RandomState, "normal", normal)
    setattr(RandomState, "pareto", pareto)
    setattr(RandomState, "poisson", poisson)
    setattr(RandomState, "power", power)
    setattr(RandomState, "rayleigh", rayleigh)
    setattr(RandomState, "standard_cauchy", standard_cauchy)
    setattr(RandomState, "standard_exponential", standard_exponential)
    setattr(RandomState, "standard_gamma", standard_gamma)
    setattr(RandomState, "standard_normal", standard_normal)
    setattr(RandomState, "standard_t", standard_t)
    setattr(RandomState, "triangular", triangular)
    setattr(RandomState, "uniform", uniform)
    setattr(RandomState, "vonmises", vonmises)
    setattr(RandomState, "wald", wald)
    setattr(RandomState, "weibull", weibull)
    setattr(RandomState, "zipf", zipf)
    setattr(RandomState, "permutation", permutation)
    setattr(RandomState, "shuffle", shuffle)


_install()
del _install


seed = _random_state.seed

rand = _random_state.rand
randn = _random_state.randn
randint = _random_state.randint
random_integers = _random_state.random_integers
random_sample = _random_state.random_sample
random = _random_state.random
ranf = _random_state.ranf
sample = _random_state.sample
choice = _random_state.choice
bytes = _random_state.bytes

permutation = _random_state.permutation
shuffle = _random_state.shuffle

beta = _random_state.beta
binomial = _random_state.binomial
chisquare = _random_state.chisquare
dirichlet = _random_state.dirichlet
exponential = _random_state.exponential
f = _random_state.f
gamma = _random_state.gamma
geometric = _random_state.geometric
gumbel = _random_state.gumbel
hypergeometric = _random_state.hypergeometric
laplace = _random_state.laplace
logistic = _random_state.logistic
lognormal = _random_state.lognormal
logseries = _random_state.logseries
multinomial = _random_state.multinomial
multivariate_normal = _random_state.multivariate_normal
negative_binomial = _random_state.negative_binomial
noncentral_chisquare = _random_state.noncentral_chisquare
noncentral_f = _random_state.noncentral_f
normal = _random_state.normal
pareto = _random_state.pareto
poisson = _random_state.poisson
power = _random_state.power
rayleigh = _random_state.rayleigh
standard_cauchy = _random_state.standard_cauchy
standard_exponential = _random_state.standard_exponential
standard_gamma = _random_state.standard_gamma
standard_normal = _random_state.standard_normal
standard_t = _random_state.standard_t
triangular = _random_state.triangular
uniform = _random_state.uniform
vonmises = _random_state.vonmises
wald = _random_state.wald
weibull = _random_state.weibull
zipf = _random_state.zipf



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas
from pandas.core.accessor import CachedAccessor
from pandas.util._decorators import doc

from ..core import Data, DataRef, DataType
from ..core.adapter import MarsDataFrame, MarsIndex, MarsSeries, to_mars
from ..core.data import register_cls_to_type
from ..core.utils.docstring import attach_module_callable_docstring
from .accessors import DatetimeAccessor, StringAccessor
from .indexing import DataFrameAt, DataFrameIat, DataFrameIloc, DataFrameLoc
from .plotting import PlotAccessor


class IndexingMixin:
    @property  # type: ignore
    @doc(DataFrameAt)
    def at(self):
        return DataFrameAt(self.data._mars_entity.at)

    @property  # type: ignore
    @doc(DataFrameIat)
    def iat(self):
        return DataFrameIat(self.data._mars_entity.iat)

    @property  # type: ignore
    @doc(DataFrameLoc)
    def loc(self):
        return DataFrameLoc(self.data._mars_entity.loc)

    @property  # type: ignore
    @doc(DataFrameIloc)
    def iloc(self):
        return DataFrameIloc(self.data._mars_entity.iloc)


@register_cls_to_type(data_type=DataType.dataframe)
class DataFrame(DataRef, IndexingMixin):
    plot = CachedAccessor("plot", PlotAccessor)

    def __init__(self, *args, **kwargs):
        data = Data(
            data_type=DataType.dataframe,
            mars_entity=MarsDataFrame(*to_mars(args), **to_mars(kwargs)),
        )
        super().__init__(data)


attach_module_callable_docstring(DataFrame, pandas, pandas.DataFrame)


@register_cls_to_type(data_type=DataType.series)
class Series(DataRef, IndexingMixin):
    str = CachedAccessor("str", StringAccessor)
    dt = CachedAccessor("dt", DatetimeAccessor)
    plot = CachedAccessor("plot", PlotAccessor)

    def __init__(self, *args, **kwargs):
        data = Data(
            data_type=DataType.series,
            mars_entity=MarsSeries(*to_mars(args), **to_mars(kwargs)),
        )
        super().__init__(data)


attach_module_callable_docstring(Series, pandas, pandas.Series)


@register_cls_to_type(data_type=DataType.index)
class Index(DataRef):
    def __init__(self, *args, **kwargs):
        data = Data(
            data_type=DataType.index,
            mars_entity=MarsIndex(*to_mars(args), **to_mars(kwargs)),
        )
        super().__init__(data)


attach_module_callable_docstring(Index, pandas, pandas.Index)



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .argpartition import argpartition
from .argsort import argsort
from .argtopk import argtopk
from .argwhere import TensorArgwhere, argwhere
from .array_split import array_split
from .astype import TensorAstype
from .atleast_1d import atleast_1d
from .atleast_2d import atleast_2d
from .atleast_3d import atleast_3d
from .broadcast_arrays import broadcast_arrays
from .broadcast_to import TensorBroadcastTo, broadcast_to
from .copy import copy
from .copyto import TensorCopyTo, copyto
from .delete import delete
from .diff import diff
from .dsplit import dsplit
from .ediff1d import ediff1d
from .expand_dims import expand_dims
from .flatten import flatten
from .flip import flip
from .fliplr import fliplr
from .flipud import flipud
from .hsplit import hsplit
from .in1d import in1d
from .insert import insert
from .isin import TensorIsIn, isin
from .map_chunk import TensorMapChunk, map_chunk
from .moveaxis import moveaxis
from .ndim import ndim
from .partition import partition
from .ravel import ravel
from .rebalance import rebalance
from .repeat import TensorRepeat, repeat
from .result_type import result_type
from .roll import roll
from .rollaxis import rollaxis
from .searchsorted import TensorSearchsorted, searchsorted
from .setdiff1d import setdiff1d
from .shape import shape
from .sort import sort
from .split import TensorSplit, split
from .squeeze import TensorSqueeze, squeeze
from .swapaxes import TensorSwapAxes, swapaxes
from .tile import tile
from .to_cpu import to_cpu
from .to_gpu import to_gpu
from .topk import topk
from .transpose import TensorTranspose, transpose
from .trapz import trapz
from .unique import unique
from .vsplit import vsplit
from .where import TensorWhere, where


def _install():
    from ..core import Tensor, TensorData
    from .astype import _astype

    for cls in (Tensor, TensorData):
        setattr(cls, "astype", _astype)
        setattr(cls, "swapaxes", swapaxes)
        setattr(cls, "squeeze", squeeze)
        setattr(cls, "repeat", repeat)
        setattr(cls, "ravel", ravel)
        setattr(cls, "flatten", flatten)
        setattr(cls, "to_gpu", to_gpu)
        setattr(cls, "to_cpu", to_cpu)
        setattr(cls, "rebalance", rebalance)
        setattr(cls, "map_chunk", map_chunk)


_install()
del _install



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import inspect
import os
import sys
import traceback
from pathlib import Path
from typing import Callable, Optional


def is_pydev_evaluating_value() -> bool:
    for frame in traceback.extract_stack():
        if "pydev" in frame.filename and frame.name == "var_to_xml":
            return True
    return False


def safe_repr_str(f: Callable):
    def inn(self, *args, **kwargs):
        if is_pydev_evaluating_value():
            # if is evaluating value from pydev, pycharm, etc
            # skip repr or str
            return getattr(object, f.__name__)(self)
        else:
            return f(self, *args, **kwargs)

    return inn


def get_local_py_version():
    """
    Get the python version on the machine where Xorbits is installed, formatted by "major.minor", like "3.10"
    """
    return str(sys.version_info.major) + "." + str(sys.version_info.minor)


def get_local_package_version(package_name: str) -> Optional[str]:
    """
    Get the version of a python package. If the package is not installed, return None
    """
    try:
        return importlib.import_module(package_name).__version__
    except ModuleNotFoundError:
        return None


def get_default_logging_config_file_path() -> Path:
    p = os.path.join(
        Path(__file__).parent.absolute(), "deploy", "oscar", "file-logging.conf"
    )
    return Path(p)


def get_non_default_kwargs(kwargs, func):
    sig = inspect.signature(func)
    defaults = {k: v.default for k, v in sig.parameters.items() if v.default != v.empty}
    dummy = object()
    kwargs = {k: v for k, v in kwargs.items() if v is not defaults.get(k, dummy)}
    for k, v in sig.parameters.items():
        if v.kind == v.VAR_KEYWORD:
            kwargs.update(kwargs.pop(k))
            break
    return kwargs


def check_signature_compatible(func1, func2, message):
    sig1 = inspect.signature(func1)
    sig2 = inspect.signature(func2)
    match_param_keys = sig1.parameters.keys() & sig2.parameters.keys()
    for k in match_param_keys:
        assert sig1.parameters[k] == sig2.parameters[k], message



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from numbers import Number
from typing import Dict, Iterable, List, Mapping, Sequence, Union

import numpy as np
from sklearn.model_selection._search import ParameterGrid as _SK_ParameterGrid

from ... import tensor as mt


class ParameterGrid(_SK_ParameterGrid):
    param_grid: List[Dict[str, Union[mt.Tensor, Sequence]]]

    def __init__(
        self, param_grid: Union[List[Dict[str, Iterable]], Mapping[str, Iterable]]
    ):
        if not isinstance(param_grid, (Mapping, Iterable)):
            raise TypeError(
                f"Parameter grid should be a dict or a list, got: {param_grid!r} of"
                f" type {type(param_grid).__name__}"
            )

        if isinstance(param_grid, Mapping):
            param_grid = [param_grid]

        for grid in param_grid:
            if not isinstance(grid, dict):
                raise TypeError(f"Parameter grid is not a dict ({grid!r})")
            for key, value in grid.items():
                if isinstance(value, (np.ndarray, Number)):
                    xnp_value = mt.array(value)
                    value = xnp_value
                    grid[key] = xnp_value
                elif isinstance(value, Iterable):
                    is_num = True
                    for i, v in enumerate(value):
                        if isinstance(v, (np.ndarray, Number)):
                            xnp_value = mt.array(v)
                            value[i] = xnp_value
                            grid[key][i] = xnp_value
                        else:
                            is_num = False
                    if is_num:
                        grid[key] = mt.ExecutableTuple(grid[key])

                if isinstance(value, (mt.Tensor)) and value.ndim > 1:
                    raise ValueError(
                        f"Parameter array for {key!r} should be one-dimensional, got:"
                        f" {value!r} with shape {value.shape}"
                    )
                if isinstance(value, str) or not isinstance(
                    value, (mt.Tensor, Sequence)
                ):
                    raise TypeError(
                        f"Parameter grid for parameter {key!r} needs to be a list or a"
                        f" mars array, but got {value!r} (of type "
                        f"{type(value).__name__}) instead. Single values "
                        "need to be wrapped in a list with one element."
                    )
                if len(value) == 0:
                    raise ValueError(
                        f"Parameter grid for parameter {key!r} need "
                        f"to be a non-empty sequence, got: {value!r}"
                    )

        self.param_grid = param_grid

    def __getitem__(self, ind):
        out = super().__getitem__(ind)
        for k, v in out.items():
            if isinstance(v, mt.Tensor):
                out[k] = v.execute()
        return out



from setuptools import setup, find_packages

setup(
    name='custom-xorbits-dg',
    version='0.1.2',
    author='antonvls',
    description='A custom version of xorbits library',
    packages=find_packages(),
    install_requires=[
   	 'numpy',
   	 'pandas',
    ],
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.10',
    ],
)




# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Any, Dict, Optional, Union

from ...._mars.core import is_build_mode
from ...._mars.core.entity import (
    OutputType,
    register_fetch_class,
    register_output_types,
)
from ...._mars.core.entity.utils import refresh_tileable_shape
from ...._mars.core.operand.objects import ObjectFetch
from ...._mars.serialization.serializables import FieldTypes, ListField
from ...dataset import Dataset, DatasetChunk, DatasetChunkData, DatasetData
from .export import export
from .getitem import getitem
from .map import map
from .rechunk import rechunk
from .to_dataframe import to_dataframe


class HuggingfaceDatasetChunkData(DatasetChunkData):
    __slots__ = ()
    type_name = "HuggingfaceDatasetChunkData"

    @classmethod
    def get_params_from_data(cls, data) -> Dict[str, Any]:
        """For updating chunk shape from data."""
        return {"shape": data.shape}


class HuggingfaceDatasetChunk(DatasetChunk):
    __slots__ = ()
    _allow_data_type_ = (HuggingfaceDatasetChunkData,)
    type_name = "HuggingfaceDatasetChunk"


class HuggingfaceDatasetData(DatasetData):
    __slots__ = ()
    type_name = "Huggingface Dataset"

    _chunks = ListField(
        "chunks",
        FieldTypes.reference(HuggingfaceDatasetChunk),
        on_serialize=lambda x: [it.data for it in x] if x is not None else x,
        on_deserialize=lambda x: [HuggingfaceDatasetChunk(it) for it in x]
        if x is not None
        else x,
    )

    def __repr__(self):
        if is_build_mode() or len(self._executed_sessions) == 0:
            # in build mode, or not executed, just return representation
            return f"Huggingface Dataset <op={type(self.op).__name__}, key={self.key}>"
        else:
            try:
                return f"Dataset({{\n    features: {self.dtypes.index.values.tolist()},\n    num_rows: {self.shape[0]}\n}})"
            except:  # noqa: E722  # nosec  # pylint: disable=bare-except  # pragma: no cover
                return (
                    f"Huggingface Dataset <op={type(self.op).__name__}, key={self.key}>"
                )

    def refresh_params(self):
        refresh_tileable_shape(self)
        # TODO(codingl2k1): update dtypes.

    def rechunk(self, num_chunks: int, **kwargs):
        return rechunk(self, num_chunks, **kwargs)

    def map(self, fn, **kwargs):
        return map(self, fn, **kwargs)

    def to_dataframe(self, types_mapper=None):
        return to_dataframe(self, types_mapper)

    def export(
        self,
        path: Union[str, os.PathLike],
        storage_options: Optional[dict] = None,
        create_if_not_exists: Optional[bool] = True,
        max_chunk_rows: Optional[int] = None,
        column_groups: Optional[dict] = None,
        num_threads: Optional[int] = None,
        version: Optional[str] = None,
        overwrite: Optional[bool] = True,
    ):
        return export(
            self,
            path,
            storage_options,
            create_if_not_exists,
            max_chunk_rows,
            column_groups,
            num_threads,
            version,
            overwrite,
        )

    def __getitem__(self, item: Union[int, slice, str]):
        return getitem(self, item)


class HuggingfaceDataset(Dataset):
    __slots__ = ()
    _allow_data_type_ = (HuggingfaceDatasetData,)
    type_name = "Huggingface Dataset"

    def to_dataset(self):
        return Dataset(self.data)


register_output_types(
    OutputType.huggingface_dataset,
    (HuggingfaceDataset, HuggingfaceDatasetData),
    (HuggingfaceDatasetChunk, HuggingfaceDatasetChunkData),
)


class HuggingfaceDatasetFetch(ObjectFetch):
    _output_type_ = OutputType.huggingface_dataset


register_fetch_class(OutputType.huggingface_dataset, HuggingfaceDatasetFetch, None)



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Dict, List

from ....core import TileableGraph, TileContext
from ....resource import Resource
from ...lifecycle.api import LifecycleAPI


def get_band_resources_from_config(
    backend_execution_config: Dict,
) -> List[Dict[str, Resource]]:
    config = backend_execution_config
    n_worker: int = config["n_worker"]
    n_cpu: int = config["n_cpu"]
    mem_bytes: int = config["mem_bytes"]
    cuda_devices: List[List[int]] = config.get("cuda_devices")

    bands_to_resource = []
    worker_cpus = n_cpu // n_worker
    cuda_devices = cuda_devices or ([[]] * n_worker)
    if sum(len(devices) for devices in cuda_devices) == 0:
        assert worker_cpus > 0, (
            f"{n_cpu} cpus are not enough " f"for {n_worker}, try to decrease workers."
        )
    mem_bytes = mem_bytes // n_worker
    for _, devices in zip(range(n_worker), cuda_devices):
        worker_band_to_resource = dict()
        if worker_cpus > 0:
            worker_band_to_resource["numa-0"] = Resource(
                num_cpus=worker_cpus, mem_bytes=mem_bytes
            )
        for i in devices:
            worker_band_to_resource[f"gpu-{i}"] = Resource(num_gpus=1)
        bands_to_resource.append(worker_band_to_resource)
    return bands_to_resource


class ResultTileablesLifecycle:
    def __init__(
        self,
        tileable_graph: TileableGraph,
        tile_context: TileContext,
        lifecycle_api: LifecycleAPI,
    ):
        self._tileable_graph = tileable_graph
        self._tile_context = tile_context
        self._lifecycle_api = lifecycle_api
        self._lifecycle_tracked_tileables = set()
        self._lifecycle_untracked_tileables = set(self._tileable_graph.result_tileables)

    async def incref_tiled(self):
        # track and incref result tileables if tiled
        tracks = [], []
        new_track_tileables = set()
        for tileable in self._lifecycle_untracked_tileables:
            try:
                tiled_tileable = self._tile_context[tileable]
            except KeyError:
                # not tiled, skip
                pass
            else:
                tileable_key = tileable.key
                tracks[0].append(tileable_key)
                tracks[1].append(
                    self._lifecycle_api.track.delay(
                        tileable_key, [c.key for c in tiled_tileable.chunks]
                    )
                )
                new_track_tileables.add(tileable)

        if any(tracks):
            # TODO(fyrestone): make the decref cancellation safe or
            # make all the tileable ids unique.
            self._lifecycle_untracked_tileables -= new_track_tileables
            self._lifecycle_tracked_tileables |= new_track_tileables
            await self._lifecycle_api.track.batch(*tracks[1])
            await self._lifecycle_api.incref_tileables(tracks[0])

    async def decref_tracked(self):
        await self._lifecycle_api.decref_tileables(
            [t.key for t in self._lifecycle_tracked_tileables]
        )



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Any, Callable, Dict, Optional, Union

import numpy as np

from .._mars.core.entity.chunks import Chunk, ChunkData
from .._mars.core.entity.tileables import HasShapeTileable, HasShapeTileableData
from .._mars.serialization.serializables import (
    FieldTypes,
    ListField,
    SeriesField,
    TupleField,
)
from .._mars.utils import on_deserialize_shape, on_serialize_shape


# DatasetChunk and DatasetChunkData can't inherit ObjectChunk and ObjectChunkData,
# because _get_output_type_by_cls() in xorbits/_mars/core/entity/output_types.py
# will generate an incorrect object output type for the DatasetChunk and DatasetChunkData.
class DatasetChunkData(ChunkData):
    __slots__ = ()
    type_name = "DatasetChunkData"

    # required for get shape of chunk, e.g. getitem.
    shape = TupleField(
        "shape",
        FieldTypes.int64,
        on_serialize=on_serialize_shape,
        on_deserialize=on_deserialize_shape,
    )

    def __init__(self, op=None, index=None, shape=None, **kwargs):
        # CheckedTaskPreprocessor._check_nsplits may check shape,
        # so the shape can't be None.
        super().__init__(
            _op=op, _index=index, shape=shape or (np.nan, np.nan), **kwargs
        )

    @property
    def params(self):
        return {
            "shape": self.shape,
            "index": self.index,
        }

    @params.setter
    def params(self, params):
        params.pop("index", None)  # index not needed to update
        self.shape = params.pop("shape", None)

    @classmethod
    def get_params_from_data(cls, data: Any) -> Dict[str, Any]:
        return dict()


class DatasetChunk(Chunk):
    __slots__ = ()
    _allow_data_type_ = (DatasetChunkData,)
    type_name = "DatasetChunk"


class DatasetData(HasShapeTileableData):
    __slots__ = ()
    type_name = "DatasetData"

    # required for to_dataframe.
    dtypes = SeriesField("dtypes")
    _chunks = ListField(
        "chunks",
        FieldTypes.reference(DatasetChunk),
        on_serialize=lambda x: [it.data for it in x] if x is not None else x,
        on_deserialize=lambda x: [DatasetChunk(it) for it in x] if x is not None else x,
    )

    def __init__(self, op=None, nsplits=None, chunks=None, shape=None, **kw):
        super().__init__(_op=op, _nsplits=nsplits, _chunks=chunks, _shape=shape, **kw)

    def __repr__(self):
        return f"Dataset <op={type(self.op).__name__}, key={self.key}>"

    @property
    def params(self):
        d = super().params
        d.update({"dtypes": self.dtypes})
        return d

    @params.setter
    def params(self, params):
        self._shape = params.pop("shape", None)
        self.dtypes = params.pop("dtypes", None)

    def refresh_params(self):
        # refresh params when chunks updated
        # nothing needs to do for Object
        pass

    def rechunk(self, num_chunks: int, **kwargs):
        raise NotImplementedError

    def map(self, fn, **kwargs):
        raise NotImplementedError

    def to_dataframe(self, types_mapper=None):
        raise NotImplementedError

    def export(
        self,
        path: Union[str, os.PathLike],
        storage_options: Optional[dict] = None,
        create_if_not_exists: Optional[bool] = True,
        max_chunk_rows: Optional[int] = None,
        column_groups: Optional[dict] = None,
        num_threads: Optional[int] = None,
        version: Optional[str] = None,
        overwrite: Optional[bool] = True,
    ):
        raise NotImplementedError

    def __getitem__(self, item):
        raise NotImplementedError


class Dataset(HasShapeTileable):
    __slots__ = ()
    _allow_data_type_ = (DatasetData,)
    type_name = "Dataset"

    def rechunk(self, num_chunks: int, **kwargs):
        """Split the internal data into chunks.

        Currently, `rechunk()` only works for single chunk dataset.

        Parameters
        ----------
        num_chunks: int
            The number of chunks.
        kwargs
            Preserved.

        Returns
        -------
            Dataset
        """
        return self.data.rechunk(num_chunks, **kwargs)

    def map(self, fn: Callable, **kwargs):
        """Apply fn to each row of the dataset.

        Parameters
        ----------
        fn: Callable
            The callable object, the signature is `function(example: Dict[str, Any]) -> Dict[str, Any]`.
        kwargs:
            The kwargs are passed to the underlying engine, e.g. the **kwargs
            will be passed to `datasets.Dataset.map` if the Dataset is constructed
            from_huggingface, please refer to:
            `datasets.Dataset.map <https://huggingface.co/docs/datasets/main/en/package_reference/main_classes#datasets.Dataset.map>`_.

        Returns
        -------
            Dataset
        Examples
        --------
        >>> import xorbits.datasets as xdatasets
        >>> ds = xdatasets.from_huggingface("rotten_tomatoes", split="validation")
        >>> def add_prefix(example):
        ...     example["text"] = "Review: " + example["text"]
        ...     return example
        >>> ds = ds.map(add_prefix)
        """
        return self.data.map(fn, **kwargs)

    def to_dataframe(self, types_mapper=None):
        """Convert the dataset to xorbits dataframe.

        The conversion will be chunk to chunk.

        Parameters
        ----------
        types_mapper: Callable
            The types mapper to pandas dataframe.

        Returns
        -------
            DataFrame
        Examples
        --------
        >>> import xorbits.datasets as xdatasets
        >>> ds = xdatasets.from_huggingface("rotten_tomatoes", split="train")
        >>> df = ds.to_dataframe()
        """
        return self.data.to_dataframe(types_mapper)

    def export(
        self,
        path: Union[str, os.PathLike],
        storage_options: Optional[dict] = None,
        create_if_not_exists: Optional[bool] = True,
        max_chunk_rows: Optional[int] = None,
        column_groups: Optional[dict] = None,
        num_threads: Optional[int] = None,
        version: Optional[str] = None,
        overwrite: Optional[bool] = True,
    ):
        """
        Export the dataset to storage.

        The storage can be local or remote, e.g. local disk or S3, ...

        Parameters
        ----------
        path: str
            The export path, can be a local path or a remote url, lease refer to:
            `fsspec <https://filesystem-spec.readthedocs.io/en/latest/intro.html>`_
        storage_options: `dict`, *optional*
            Key/value pairs to be passed on to the caching file-system backend, if any.
        create_if_not_exists: bool
            Whether to create the path if it does not exist.
        max_chunk_rows: int
            Max rows per chunk file, default is 100.
        column_groups: dict
            A dict of group name string to a list of column index or name.
        num_threads: int
            The thread concurrency on each chunk.
        version: str
            The version string, default is 0.0.0.
        overwrite: bool
            Whether overwrites the dataset version.

        Returns
        -------
            A dict of export info.
        Examples
        --------

        Export to local disk.

        >>> import xorbits.datasets as xdatasets
        >>> ds = xdatasets.from_huggingface("cifar10", split="train")
        >>> ds.export("./export_dir")

        Export to remote storage.

        >>> import xorbits.datasets as xdatasets
        >>> storage_options = {"key": aws_access_key_id, "secret": aws_secret_access_key}
        >>> ds = xdatasets.from_huggingface("cifar10", split="train")
        >>> ds.export("./export_dir", storage_options=storage_options)

        """
        return self.data.export(
            path,
            storage_options,
            create_if_not_exists,
            max_chunk_rows,
            column_groups,
            num_threads,
            version,
            overwrite,
        )

    def __getitem__(self, item: Union[int, slice, str]):
        """Get rows or columns from dataset.

        The result will be formatted dict or list.

        Parameters
        ----------
        item: Union[int, slice, str]
            The item index. slice does not support steps, e.g. ds[1:3:2] is not supported.

        Returns
        -------
            List of column values if item is str.
            Dict[column name, List[values of selected rows]] if item is int or slice.
        Examples
        --------
        >>> import xorbits.datasets as xdatasets
        >>> ds = xdatasets.from_huggingface("rotten_tomatoes", split="train")
        >>> ds[1:3]["text"]
        """
        return self.data.__getitem__(item)



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import scipy.special as spspecial

from ... import opcodes
from ...core import ExecutableTuple
from ...utils import is_same_module
from ..arithmetic.core import TensorBinOp, TensorMultiOp, TensorUnaryOp
from ..array_utils import (
    as_same_device,
    convert_order,
    cp,
    device,
    issparse,
    np,
    sparse,
)
from ..datasource import tensor as astensor

_func_name_to_special_cls = {}


def _register_special_op(cls):
    if cls._func_name is not None:
        _func_name_to_special_cls[cls._func_name] = cls
    return cls


class TensorSpecialOperandMixin:
    _op_code_ = opcodes.SPECIAL
    _func_name = None

    def __new__(cls, *args, **kwargs):
        if cls._func_name is not None:
            return object.__new__(_func_name_to_special_cls[cls._func_name])
        return super().__new__(cls, *args, **kwargs)

    @classmethod
    def _get_func(cls, xp):
        if is_same_module(xp, np):
            from scipy import special

            return getattr(special, cls._func_name)
        elif cp is not None and is_same_module(xp, cp):
            from cupyx.scipy import special

            return getattr(special, cls._func_name)
        else:
            assert is_same_module(xp, sparse)
            return getattr(sparse, cls._func_name)


class TensorSpecialUnaryOp(TensorSpecialOperandMixin, TensorUnaryOp):
    pass


class TensorSpecialBinOp(TensorSpecialOperandMixin, TensorBinOp):
    pass


class TensorSpecialMultiOp(TensorSpecialOperandMixin, TensorMultiOp):
    @classmethod
    def _execute_gpu(cls, op, xp, *args, **kw):
        if kw.get("out") is not None:
            kw["out"] = xp.asarray(kw["out"])
        r = cls._get_func(xp)(*args, **kw)
        return convert_order(r, op.outputs[0].order.value)

    @classmethod
    def _execute_cpu(cls, op, xp, *args, **kw):
        kw["order"] = op.order
        if kw.get("out") is not None:
            kw["out"] = np.asarray(kw["out"])
        try:
            return cls._get_func(xp)(*args, **kw)
        except TypeError:
            kw.pop("order")
            r = cls._get_func(xp)(*args, **kw)
            if issparse(r):
                return r
            return convert_order(r, op.outputs[0].order.value)

    @classmethod
    def execute(cls, ctx, op):
        inputs, device_id, xp = as_same_device(
            [ctx[c.key] for c in op.inputs], device=op.device, ret_extra=True
        )

        with device(device_id):
            kw = {"casting": op.casting} if op.out is not None else {}

            inputs_iter = iter(inputs)
            args = [a if np.isscalar(a) else next(inputs_iter) for a in op.args]
            if op.out is not None:
                kw["out"] = next(inputs_iter).copy()

            with np.errstate(**op.err):
                if op.is_gpu():
                    ret = cls._execute_gpu(op, xp, *args, **kw)
                else:
                    ret = cls._execute_cpu(op, xp, *args, **kw)

                if ret.dtype != op.dtype:
                    ret = ret.astype(op.dtype)
                ctx[op.outputs[0].key] = ret


class TensorTupleOp(TensorSpecialUnaryOp):
    @property
    def output_limit(self):
        return self._n_outputs

    def __call__(self, x, out=None):
        x = astensor(x)

        if out is not None:
            if not isinstance(out, ExecutableTuple):
                raise TypeError(
                    f"out should be ExecutableTuple object, got {type(out)} instead"
                )
            if len(out) != self._n_outputs:
                raise TypeError(
                    f"out should be an ExecutableTuple object with {self._n_outputs} elements, got {len(out)} instead"
                )

        func = getattr(spspecial, self._func_name)
        res = func(np.ones(x.shape, dtype=x.dtype))
        res_tensors = self.new_tensors(
            [x],
            kws=[
                {
                    "side": f"{self._func_name}[{i}]",
                    "dtype": output.dtype,
                    "shape": output.shape,
                }
                for i, output in enumerate(res)
            ],
        )

        if out is None:
            return ExecutableTuple(res_tensors)

        for res_tensor, out_tensor in zip(res_tensors, out):
            out_tensor.data = res_tensor.data
        return out

    @classmethod
    def execute(cls, ctx, op):
        inputs, device_id, xp = as_same_device(
            [ctx[c.key] for c in op.inputs], device=op.device, ret_extra=True
        )

        with device(device_id):
            with np.errstate(**op.err):
                if op.is_gpu():
                    ret = cls._execute_gpu(op, xp, inputs[0])
                else:
                    ret = cls._execute_cpu(op, xp, inputs[0])

                for output, ret_element in zip(op.outputs, ret):
                    ctx[output.key] = ret_element



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...core import is_build_mode
from .abs import TensorAbs, abs
from .absolute import TensorAbsolute, absolute
from .add import TensorAdd, TensorTreeAdd, add, tree_add
from .angle import TensorAngle, angle
from .arccos import TensorArccos, arccos
from .arccosh import TensorArccosh, arccosh
from .arcsin import TensorArcsin, arcsin
from .arcsinh import TensorArcsinh, arcsinh
from .arctan import TensorArctan, arctan
from .arctan2 import TensorArctan2, arctan2
from .arctanh import TensorArctanh, arctanh
from .around import TensorAround
from .around import around
from .around import around as round_
from .bitand import TensorBitand, bitand
from .bitor import TensorBitor, bitor
from .bitxor import TensorBitxor, bitxor
from .cbrt import TensorCbrt, cbrt
from .ceil import TensorCeil, ceil
from .clip import TensorClip, clip
from .conj import TensorConj
from .conj import conj
from .conj import conj as conjugate
from .copysign import TensorCopysign, copysign
from .cos import TensorCos, cos
from .cosh import TensorCosh, cosh
from .deg2rad import TensorDeg2rad, deg2rad
from .degrees import TensorDegrees, degrees
from .divide import TensorDivide, divide
from .equal import TensorEqual, equal
from .exp import TensorExp, exp
from .exp2 import TensorExp2, exp2
from .expm1 import TensorExpm1, expm1
from .fabs import TensorFabs, fabs
from .fix import TensorFix, fix
from .float_power import TensorFloatPower, float_power
from .floor import TensorFloor, floor
from .floordiv import TensorFloorDiv, floordiv
from .fmax import TensorFMax, fmax
from .fmin import TensorFMin, fmin
from .fmod import TensorFMod, fmod
from .frexp import TensorFrexp, frexp
from .greater import TensorGreaterThan, greater
from .greater_equal import TensorGreaterEqual, greater_equal
from .hypot import TensorHypot, hypot
from .i0 import TensorI0, i0
from .imag import TensorImag, imag
from .invert import TensorInvert, invert
from .isclose import TensorIsclose, isclose
from .iscomplex import TensorIsComplex, iscomplex
from .isfinite import TensorIsFinite, isfinite
from .isinf import TensorIsInf, isinf
from .isnan import TensorIsNan, isnan
from .isreal import TensorIsReal, isreal
from .ldexp import TensorLdexp, ldexp
from .less import TensorLessThan, less
from .less_equal import TensorLessEqual, less_equal
from .log import TensorLog, log
from .log1p import TensorLog1p, log1p
from .log2 import TensorLog2, log2
from .log10 import TensorLog10, log10
from .logaddexp import TensorLogAddExp, logaddexp
from .logaddexp2 import TensorLogAddExp2, logaddexp2
from .logical_and import TensorAnd, logical_and
from .logical_not import TensorNot, logical_not
from .logical_or import TensorOr, logical_or
from .logical_xor import TensorXor, logical_xor
from .lshift import TensorLshift, lshift
from .maximum import TensorMaximum, maximum
from .minimum import TensorMinimum, minimum
from .mod import TensorMod
from .mod import mod
from .mod import mod as remainder
from .modf import TensorModf, modf
from .multiply import TensorMultiply, TensorTreeMultiply, multiply, tree_multiply
from .nan_to_num import TensorNanToNum, nan_to_num
from .negative import TensorNegative, negative
from .nextafter import TensorNextafter, nextafter
from .not_equal import TensorNotEqual, not_equal
from .positive import TensorPositive, positive
from .power import TensorPower, power
from .rad2deg import TensorRad2deg, rad2deg
from .radians import TensorRadians, radians
from .real import TensorReal, real
from .reciprocal import TensorReciprocal, reciprocal
from .rint import TensorRint, rint
from .rshift import TensorRshift, rshift
from .setimag import TensorSetImag
from .setreal import TensorSetReal
from .sign import TensorSign, sign
from .signbit import TensorSignbit, signbit
from .sin import TensorSin, sin
from .sinc import TensorSinc, sinc
from .sinh import TensorSinh, sinh
from .spacing import TensorSpacing, spacing
from .sqrt import TensorSqrt, sqrt
from .square import TensorSquare, square
from .subtract import TensorSubtract, subtract
from .tan import TensorTan, tan
from .tanh import TensorTanh, tanh
from .truediv import TensorTrueDiv, truediv
from .trunc import TensorTrunc, trunc


def _wrap_iop(func):
    def inner(self, *args, **kwargs):
        kwargs["out"] = self
        return func(self, *args, **kwargs)

    return inner


def _install():
    from ..core import TENSOR_TYPE, Tensor, TensorData
    from ..datasource import tensor as astensor
    from .add import add, radd
    from .bitand import bitand, rbitand
    from .bitor import bitor, rbitor
    from .bitxor import bitxor, rbitxor
    from .divide import divide, rdivide
    from .floordiv import floordiv, rfloordiv
    from .lshift import lshift, rlshift
    from .mod import mod, rmod
    from .multiply import multiply, rmultiply
    from .power import power, rpower
    from .rshift import rrshift, rshift
    from .subtract import rsubtract, subtract
    from .truediv import rtruediv, truediv

    def _wrap_equal(func):
        def eq(x1, x2, **kwargs):
            if is_build_mode():
                return astensor(x1)._equals(x2)
            return func(x1, x2, **kwargs)

        return eq

    for cls in TENSOR_TYPE:
        setattr(cls, "__add__", add)
        setattr(cls, "__iadd__", _wrap_iop(add))
        setattr(cls, "__radd__", radd)
        setattr(cls, "__sub__", subtract)
        setattr(cls, "__isub__", _wrap_iop(subtract))
        setattr(cls, "__rsub__", rsubtract)
        setattr(cls, "__mul__", multiply)
        setattr(cls, "__imul__", _wrap_iop(multiply))
        setattr(cls, "__rmul__", rmultiply)
        setattr(cls, "__div__", divide)
        setattr(cls, "__idiv__", _wrap_iop(divide))
        setattr(cls, "__rdiv__", rdivide)
        setattr(cls, "__truediv__", truediv)
        setattr(cls, "__itruediv__", _wrap_iop(truediv))
        setattr(cls, "__rtruediv__", rtruediv)
        setattr(cls, "__floordiv__", floordiv)
        setattr(cls, "__ifloordiv__", _wrap_iop(floordiv))
        setattr(cls, "__rfloordiv__", rfloordiv)
        setattr(cls, "__pow__", power)
        setattr(cls, "__ipow__", _wrap_iop(power))
        setattr(cls, "__rpow__", rpower)
        setattr(cls, "__mod__", mod)
        setattr(cls, "__imod__", _wrap_iop(mod))
        setattr(cls, "__rmod__", rmod)
        setattr(cls, "__lshift__", lshift)
        setattr(cls, "__ilshift__", _wrap_iop(lshift))
        setattr(cls, "__rlshift__", rlshift)
        setattr(cls, "__rshift__", rshift)
        setattr(cls, "__irshift__", _wrap_iop(rshift))
        setattr(cls, "__rrshift__", rrshift)

        setattr(cls, "__eq__", _wrap_equal(equal))
        setattr(cls, "__ne__", not_equal)
        setattr(cls, "__lt__", less)
        setattr(cls, "__le__", less_equal)
        setattr(cls, "__gt__", greater)
        setattr(cls, "__ge__", greater_equal)
        setattr(cls, "__and__", bitand)
        setattr(cls, "__iand__", _wrap_iop(bitand))
        setattr(cls, "__rand__", rbitand)
        setattr(cls, "__or__", bitor)
        setattr(cls, "__ior__", _wrap_iop(bitor))
        setattr(cls, "__ror__", rbitor)
        setattr(cls, "__xor__", bitxor)
        setattr(cls, "__ixor__", _wrap_iop(bitxor))
        setattr(cls, "__rxor__", rbitxor)

        setattr(cls, "__neg__", negative)
        setattr(cls, "__pos__", positive)
        setattr(cls, "__abs__", abs)
        setattr(cls, "__invert__", invert)

    setattr(Tensor, "round", round_)
    setattr(Tensor, "conj", conj)
    setattr(Tensor, "conjugate", conjugate)
    setattr(TensorData, "round", round_)
    setattr(TensorData, "conj", conj)
    setattr(TensorData, "conjugate", conjugate)


_install()
del _install


BIN_UFUNC = {
    add,
    subtract,
    multiply,
    divide,
    truediv,
    floordiv,
    power,
    mod,
    fmod,
    logaddexp,
    logaddexp2,
    equal,
    not_equal,
    less,
    less_equal,
    greater,
    greater_equal,
    arctan2,
    hypot,
    bitand,
    bitor,
    bitxor,
    lshift,
    rshift,
    logical_and,
    logical_or,
    logical_xor,
    maximum,
    minimum,
    float_power,
    remainder,
    fmax,
    fmin,
    copysign,
    nextafter,
    ldexp,
}

UNARY_UFUNC = {
    square,
    arcsinh,
    rint,
    sign,
    conj,
    tan,
    absolute,
    deg2rad,
    log,
    fabs,
    exp2,
    invert,
    negative,
    sqrt,
    arctan,
    positive,
    cbrt,
    log10,
    sin,
    rad2deg,
    log2,
    arcsin,
    expm1,
    arctanh,
    cosh,
    sinh,
    cos,
    reciprocal,
    tanh,
    log1p,
    exp,
    arccos,
    arccosh,
    around,
    logical_not,
    conjugate,
    isfinite,
    isinf,
    isnan,
    signbit,
    spacing,
    floor,
    ceil,
    trunc,
    degrees,
    radians,
    angle,
    isreal,
    iscomplex,
    real,
    imag,
    fix,
    i0,
    sinc,
    nan_to_num,
}



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import inspect
from types import ModuleType
from typing import Any, Callable, Dict, List, Optional, Set

import numpy
import scipy

from ...core.adapter import MARS_TENSOR_TYPE, mars_tensor, wrap_mars_callable


def _collect_module_callables(
    m: ModuleType,
    docstring_src_module: ModuleType,
    skip_members: Optional[List[str]] = None,
) -> Dict[str, Callable]:
    module_callables: Dict[str, Callable] = dict()

    # install module functions.
    for name, func in inspect.getmembers(m, callable):
        if skip_members is not None and name in skip_members:
            continue
        module_callables[name] = wrap_mars_callable(
            func,
            attach_docstring=True,
            is_cls_member=False,
            docstring_src_module=docstring_src_module,
            docstring_src=getattr(docstring_src_module, name, None),
        )

    return module_callables


MARS_TENSOR_CALLABLES: Dict[str, Callable] = _collect_module_callables(
    mars_tensor, numpy, skip_members=["bool", "float", "int", "object"]
)
MARS_TENSOR_RANDOM_CALLABLES: Dict[str, Callable] = _collect_module_callables(
    mars_tensor.random, numpy.random
)
MARS_TENSOR_FFT_CALLABLES: Dict[str, Callable] = _collect_module_callables(
    mars_tensor.fft, numpy.fft
)
MARS_TENSOR_LINALG_CALLABLES: Dict[str, Callable] = _collect_module_callables(
    mars_tensor.linalg, numpy.linalg
)
MARS_TENSOR_SPECIAL_CALLABLES: Dict[str, Callable] = _collect_module_callables(
    mars_tensor.special, scipy.special
)


def _collect_tensor_objects():
    from ..lib.index_tricks import c_, mgrid, ogrid, r_

    mars_tensor_objects: Dict[str, Any] = dict()
    mars_tensor_objects["mgrid"] = mgrid
    mars_tensor_objects["ogrid"] = ogrid
    mars_tensor_objects["c_"] = c_
    mars_tensor_objects["r_"] = r_
    return mars_tensor_objects


MARS_TENSOR_OBJECTS: Dict[str, Any] = _collect_tensor_objects()


def _collect_tensor_magic_methods() -> Set[str]:
    magic_methods: Set[str] = set()

    magic_methods_to_skip: Set[str] = {
        "__init__",
        "__dir__",
        "__repr__",
        "__str__",
        "__setattr__",
        "__getattr__",
        "__len__",
        "__array__",
    }

    for mars_cls in MARS_TENSOR_TYPE:
        for name, _ in inspect.getmembers(mars_cls, inspect.isfunction):
            if (
                name.startswith("__")
                and name.endswith("__")
                and name not in magic_methods_to_skip
            ):
                magic_methods.add(name)
    return magic_methods


MARS_TENSOR_MAGIC_METHODS: Set[str] = _collect_tensor_magic_methods()
